# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import typing
from typing import Optional, Union
from urllib.parse import urljoin

import requests
from nuclio.auth import AuthInfo as NuclioAuthInfo
from nuclio.auth import AuthKinds as NuclioAuthKinds

import mlrun
import mlrun.common.constants as mlrun_constants
import mlrun.common.helpers
import mlrun.common.schemas as schemas
import mlrun.common.types
from mlrun.model import ModelObj
from mlrun.platforms.iguazio import min_iguazio_versions
from mlrun.utils import logger

from .function import min_nuclio_versions


class Authenticator(typing.Protocol):
    @property
    def authentication_mode(self) -> str:
        return schemas.APIGatewayAuthenticationMode.none.value

    @classmethod
    def from_scheme(cls, api_gateway_spec: schemas.APIGatewaySpec):
        if (
            api_gateway_spec.authenticationMode
            == schemas.APIGatewayAuthenticationMode.basic.value
        ):
            if api_gateway_spec.authentication:
                return BasicAuth(
                    username=api_gateway_spec.authentication.get("username", ""),
                    password=api_gateway_spec.authentication.get("password", ""),
                )
            else:
                return BasicAuth()
        elif (
            api_gateway_spec.authenticationMode
            == schemas.APIGatewayAuthenticationMode.access_key.value
        ):
            return AccessKeyAuth()
        else:
            return NoneAuth()

    def to_scheme(
        self,
    ) -> Optional[dict[str, Optional[schemas.APIGatewayBasicAuth]]]:
        return None


class APIGatewayAuthenticator(Authenticator, ModelObj):
    _dict_fields = ["authentication_mode"]


class NoneAuth(APIGatewayAuthenticator):
    """
    An API gateway authenticator with no authentication.
    """

    pass


class BasicAuth(APIGatewayAuthenticator):
    """
    An API gateway authenticator with basic authentication.

    :param username: (str) The username for basic authentication.
    :param password: (str) The password for basic authentication.
    """

    def __init__(self, username=None, password=None):
        self._username = username
        self._password = password

    @property
    def authentication_mode(self) -> str:
        return schemas.APIGatewayAuthenticationMode.basic.value

    def to_scheme(
        self,
    ) -> Optional[dict[str, Optional[schemas.APIGatewayBasicAuth]]]:
        return {
            "basicAuth": schemas.APIGatewayBasicAuth(
                username=self._username, password=self._password
            )
        }


class AccessKeyAuth(APIGatewayAuthenticator):
    """
    An API gateway authenticator with access key authentication.
    """

    @property
    def authentication_mode(self) -> str:
        return schemas.APIGatewayAuthenticationMode.access_key.value


class APIGatewayMetadata(ModelObj):
    _dict_fields = ["name", "namespace", "labels", "annotations", "creation_timestamp"]

    def __init__(
        self,
        name: str,
        namespace: Optional[str] = None,
        labels: Optional[dict] = None,
        annotations: Optional[dict] = None,
        creation_timestamp: Optional[str] = None,
    ):
        """
        :param name: The name of the API gateway
        :param namespace: The namespace of the API gateway
        :param labels: The labels of the API gateway
        :param annotations: The annotations of the API gateway
        :param creation_timestamp: The creation timestamp of the API gateway
        """
        self.name = name
        self.namespace = namespace
        self.labels = labels or {}
        self.annotations = annotations or {}
        self.creation_timestamp = creation_timestamp

        if not self.name:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "API Gateway name cannot be empty"
            )


class APIGatewaySpec(ModelObj):
    _dict_fields = [
        "functions",
        "project",
        "name",
        "description",
        "host",
        "path",
        "authentication",
        "canary",
    ]

    def __init__(
        self,
        functions: Union[
            list[
                Union[
                    str,
                    "mlrun.runtimes.nuclio.function.RemoteRuntime",
                    "mlrun.runtimes.nuclio.serving.ServingRuntime",
                    "mlrun.runtimes.nuclio.application.ApplicationRuntime",
                ]
            ],
            "mlrun.runtimes.nuclio.function.RemoteRuntime",
            "mlrun.runtimes.nuclio.serving.ServingRuntime",
            "mlrun.runtimes.nuclio.application.ApplicationRuntime",
        ],
        project: Optional[str] = None,
        description: str = "",
        host: Optional[str] = None,
        path: str = "/",
        authentication: Optional[APIGatewayAuthenticator] = NoneAuth(),
        canary: Optional[list[int]] = None,
        ports: Optional[list[int]] = None,
    ):
        """
        :param functions: The list of functions associated with the API gateway
            Can be a list of function names (["my-func1", "my-func2"])
            or a list or a single entity of
            :py:class:`~mlrun.runtimes.nuclio.function.RemoteRuntime` OR
            :py:class:`~mlrun.runtimes.nuclio.serving.ServingRuntime` OR
            :py:class:`~mlrun.runtimes.nuclio.application.ApplicationRuntime`
        :param project: The project name
        :param description: Optional description of the API gateway
        :param path: Optional path of the API gateway, default value is "/"
        :param authentication: The authentication for the API gateway of type
                :py:class:`~mlrun.runtimes.nuclio.api_gateway.BasicAuth`
        :param host:  The host of the API gateway (optional). If not set, it will be automatically generated
        :param canary: The canary percents for the API gateway of type list[int]; for instance: [20,80] (optional)
        :param ports: The ports of the API gateway, as a list of integers that correspond to the functions in the
            functions list. for instance: [8050] or [8050, 8081] (optional)
        """
        self.description = description
        self.host = host
        self.path = path
        self.authentication = authentication
        self.functions = functions
        self.canary = canary
        self.project = project
        self.ports = ports

        self.enrich()
        self.validate(project=project, functions=functions, canary=canary, ports=ports)

    def enrich(self):
        if self.path and not self.path.startswith("/"):
            self.path = f"/{self.path}"

    def validate(
        self,
        project: str,
        functions: Union[
            list[
                Union[
                    str,
                    "mlrun.runtimes.nuclio.function.RemoteRuntime",
                    "mlrun.runtimes.nuclio.serving.ServingRuntime",
                    "mlrun.runtimes.nuclio.application.ApplicationRuntime",
                ]
            ],
            "mlrun.runtimes.nuclio.function.RemoteRuntime",
            "mlrun.runtimes.nuclio.serving.ServingRuntime",
            "mlrun.runtimes.nuclio.application.ApplicationRuntime",
        ],
        canary: Optional[list[int]] = None,
        ports: Optional[list[int]] = None,
    ):
        self.functions = self._validate_functions(project=project, functions=functions)

        # validating canary
        if canary:
            self.canary = self._validate_canary(canary)

        # validating ports
        if ports:
            self.ports = self._validate_ports(ports)

    def _validate_canary(self, canary: list[int]):
        if len(self.functions) != len(canary):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Function and canary lists lengths do not match"
            )
        for canary_percent in canary:
            if canary_percent < 0 or canary_percent > 100:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "The percentage value must be in the range from 0 to 100"
                )
        if sum(canary) != 100:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "The sum of canary function percents should be equal to 100"
            )
        return canary

    def _validate_ports(self, ports):
        if len(self.functions) != len(ports):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Function and port lists lengths do not match"
            )

        return ports

    @staticmethod
    def _validate_functions(
        project: str,
        functions: Union[
            list[
                Union[
                    str,
                    "mlrun.runtimes.nuclio.function.RemoteRuntime",
                    "mlrun.runtimes.nuclio.serving.ServingRuntime",
                    "mlrun.runtimes.nuclio.application.ApplicationRuntime",
                ]
            ],
            "mlrun.runtimes.nuclio.function.RemoteRuntime",
            "mlrun.runtimes.nuclio.serving.ServingRuntime",
            "mlrun.runtimes.nuclio.application.ApplicationRuntime",
        ],
    ):
        if not isinstance(functions, list):
            functions = [functions]

        # validating functions
        if not 1 <= len(functions) <= 2:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Gateway can be created from one or two functions, "
                f"the number of functions passed is {len(functions)}"
            )

        function_names = []
        for func in functions:
            if isinstance(func, str):
                # check whether the function was passed as a URI or just a name
                parsed_project, function_name, _, _ = (
                    mlrun.common.helpers.parse_versioned_object_uri(func)
                )

                if parsed_project and function_name:
                    # check that parsed project and passed project are the same
                    if parsed_project != project:
                        raise mlrun.errors.MLRunInvalidArgumentError(
                            "Function doesn't belong to passed project"
                        )
                    function_uri = func
                else:
                    function_uri = mlrun.utils.generate_object_uri(project, func)
                function_names.append(function_uri)
                continue

            function_name = (
                func.metadata.name if hasattr(func, "metadata") else func.name
            )
            if func.kind not in mlrun.runtimes.RuntimeKinds.nuclio_runtimes():
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Input function {function_name} is not a Nuclio function"
                )
            if func.metadata.project != project:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"input function {function_name} "
                    f"does not belong to this project"
                )
            function_uri = mlrun.utils.generate_object_uri(
                project,
                function_name,
                func.metadata.tag,
                func.metadata.hash,
            )
            function_names.append(function_uri)
        return function_names


class APIGatewayStatus(ModelObj):
    def __init__(self, state: Optional[schemas.APIGatewayState] = None):
        self.state = state or schemas.APIGatewayState.none


class APIGateway(ModelObj):
    _dict_fields = [
        "metadata",
        "spec",
        "state",
    ]

    @min_nuclio_versions("1.13.1")
    def __init__(
        self,
        metadata: APIGatewayMetadata,
        spec: APIGatewaySpec,
        status: Optional[APIGatewayStatus] = None,
    ):
        """
        Initialize the APIGateway instance.

        :param metadata: (APIGatewayMetadata) The metadata of the API gateway.
        :param spec: (APIGatewaySpec) The spec of the API gateway.
        :param status: (APIGatewayStatus) The status of the API gateway.
        """
        self.metadata = metadata
        self.spec = spec
        self.status = status

    @property
    def metadata(self) -> APIGatewayMetadata:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = self._verify_dict(metadata, "metadata", APIGatewayMetadata)

    @property
    def spec(self) -> APIGatewaySpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", APIGatewaySpec)

    @property
    def status(self) -> APIGatewayStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", APIGatewayStatus)

    def invoke(
        self,
        method="POST",
        headers: Optional[dict] = None,
        credentials: Optional[tuple[str, str]] = None,
        path: Optional[str] = None,
        body: Optional[Union[str, bytes, dict]] = None,
        **kwargs,
    ):
        """
        Invoke the API gateway.

        :param method: (str, optional) The HTTP method for the invocation.
        :param headers: (dict, optional) The HTTP headers for the invocation.
        :param credentials: (Optional[tuple[str, str]], optional) The (username,password) for the invocation if required
            can also be set by the environment variable (_, V3IO_ACCESS_KEY) for access key authentication.
        :param path: (str, optional) The sub-path for the invocation.
        :param body: (Optional[Union[str, bytes, dict]]) The body of the invocation.
        :param kwargs: (dict) Additional keyword arguments.

        :return: The response from the API gateway invocation.
        """
        if not self.invoke_url:
            # try to resolve invoke_url before fail
            self.sync()
            if not self.invoke_url:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Invocation url is not set. Set up gateway's `invoke_url` attribute."
                )
        if not self.is_ready():
            raise mlrun.errors.MLRunPreconditionFailedError(
                f"API gateway is not ready. " f"Current state: {self.status.state}"
            )

        auth = None

        if (
            self.spec.authentication.authentication_mode
            == schemas.APIGatewayAuthenticationMode.basic.value
        ):
            if not credentials:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "API Gateway invocation requires authentication. Please pass credentials"
                )
            auth = NuclioAuthInfo(
                username=credentials[0], password=credentials[1]
            ).to_requests_auth()

        if (
            self.spec.authentication.authentication_mode
            == schemas.APIGatewayAuthenticationMode.access_key.value
        ):
            # inject access key from env
            if credentials:
                auth = NuclioAuthInfo(
                    username=credentials[0],
                    password=credentials[1],
                    mode=NuclioAuthKinds.iguazio,
                ).to_requests_auth()
            else:
                auth = NuclioAuthInfo().from_envvar().to_requests_auth()
            if not auth:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "API Gateway invocation requires authentication. Please set V3IO_ACCESS_KEY env var"
                )
        url = urljoin(self.invoke_url, path or "")

        # Determine the correct keyword argument for the body
        if isinstance(body, dict):
            kwargs["json"] = body
        elif isinstance(body, (str, bytes)):
            kwargs["data"] = body

        return requests.request(
            method=method,
            url=url,
            headers=headers or {},
            auth=auth,
            **kwargs,
        )

    def wait_for_readiness(self, max_wait_time=90):
        """
        Wait for the API gateway to become ready within the maximum wait time.

        Parameters:
            max_wait_time: int - Maximum time to wait in seconds (default is 90 seconds).

        Returns:
            bool: True if the entity becomes ready within the maximum wait time, False otherwise
        """

        def _ensure_ready():
            if not self.is_ready():
                raise AssertionError(
                    f"Waiting for gateway readiness is taking more than {max_wait_time} seconds"
                )

        return mlrun.utils.helpers.retry_until_successful(
            3, max_wait_time, logger, False, _ensure_ready
        )

    def is_ready(self):
        if self.status.state is not schemas.api_gateway.APIGatewayState.ready:
            # try to sync the state
            self.sync()
        return self.status.state == schemas.api_gateway.APIGatewayState.ready

    def sync(self):
        """
        Synchronize the API gateway from the server.
        """
        synced_gateway = mlrun.get_run_db().get_api_gateway(
            self.metadata.name, self.spec.project
        )
        synced_gateway = self.from_scheme(synced_gateway)

        self.spec.host = synced_gateway.spec.host
        self.spec.path = synced_gateway.spec.path
        self.spec.authentication = synced_gateway.spec.authentication
        self.spec.functions = synced_gateway.spec.functions
        self.spec.canary = synced_gateway.spec.canary
        self.spec.description = synced_gateway.spec.description
        self.status.state = synced_gateway.status.state

    def with_basic_auth(self, username: str, password: str):
        """
        Set basic authentication for the API gateway.

        :param username: (str) The username for basic authentication.
        :param password: (str) The password for basic authentication.
        """
        self.spec.authentication = BasicAuth(username=username, password=password)

    @min_iguazio_versions("3.5.5")
    def with_access_key_auth(self):
        """
        Set access key authentication for the API gateway.
        """
        self.spec.authentication = AccessKeyAuth()

    def with_canary(
        self,
        functions: Union[
            list[
                Union[
                    str,
                    "mlrun.runtimes.nuclio.function.RemoteRuntime",
                    "mlrun.runtimes.nuclio.serving.ServingRuntime",
                    "mlrun.runtimes.nuclio.application.ApplicationRuntime",
                ]
            ],
            "mlrun.runtimes.nuclio.function.RemoteRuntime",
            "mlrun.runtimes.nuclio.serving.ServingRuntime",
            "mlrun.runtimes.nuclio.application.ApplicationRuntime",
        ],
        canary: list[int],
    ):
        """
        Set canary function for the API gateway

        :param functions: The list of functions associated with the API gateway
            Can be a list of function names (["my-func1", "my-func2"])
            or a list of nuclio functions of types
            :py:class:`~mlrun.runtimes.nuclio.function.RemoteRuntime` OR
            :py:class:`~mlrun.runtimes.nuclio.serving.ServingRuntime` OR
            :py:class:`~mlrun.runtimes.nuclio.application.ApplicationRuntime`
        :param canary: The canary percents for the API gateway of type list[int]; for instance: [20,80]

        """
        if len(functions) != 2:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Gateway with canary can be created only with two functions, "
                f"the number of functions passed is {len(functions)}"
            )
        self.spec.validate(
            project=self.spec.project, functions=functions, canary=canary
        )

    def with_ports(self, ports: list[int]):
        """
        Set ports for the API gateway

        :param ports: The ports of the API gateway, as a list of integers that correspond to the functions in the
            functions list. for instance: [8050] or [8050, 8081]
        """
        self.spec.validate(
            project=self.spec.project, functions=self.spec.functions, ports=ports
        )

    def with_force_ssl_redirect(self):
        """
        Set SSL redirect annotation for the API gateway.
        """
        self.metadata.annotations["nginx.ingress.kubernetes.io/force-ssl-redirect"] = (
            "true"
        )

    def with_gateway_timeout(self, gateway_timeout: int):
        """
        Set gateway proxy connect/read/send timeout annotations
        :param gateway_timeout: The timeout in seconds
        """
        mlrun.runtimes.utils.enrich_gateway_timeout_annotations(
            self.metadata.annotations, gateway_timeout
        )

    def with_annotations(self, annotations: dict):
        """set a key/value annotations in the metadata of the api gateway"""
        for key, value in annotations.items():
            self.metadata.annotations[key] = str(value)
        return self

    @classmethod
    def from_scheme(cls, api_gateway: schemas.APIGateway):
        project = api_gateway.metadata.labels.get(
            mlrun_constants.MLRunInternalLabels.nuclio_project_name
        )
        functions, canary = cls._resolve_canary(api_gateway.spec.upstreams)
        state = (
            api_gateway.status.state
            if api_gateway.status
            else schemas.APIGatewayState.none
        )
        new_api_gateway = cls(
            metadata=APIGatewayMetadata(
                name=api_gateway.spec.name,
                annotations=api_gateway.metadata.annotations,
                labels=api_gateway.metadata.labels,
            ),
            spec=APIGatewaySpec(
                project=project,
                description=api_gateway.spec.description,
                host=api_gateway.spec.host,
                path=api_gateway.spec.path,
                authentication=APIGatewayAuthenticator.from_scheme(api_gateway.spec),
                functions=functions,
                canary=canary,
            ),
            status=APIGatewayStatus(state=state),
        )
        return new_api_gateway

    def to_scheme(self) -> schemas.APIGateway:
        upstreams = (
            [
                schemas.APIGatewayUpstream(
                    nucliofunction={"name": self.spec.functions[0]},
                    percentage=self.spec.canary[0],
                ),
                schemas.APIGatewayUpstream(
                    # do not set percent for the second function,
                    # so we can define which function to display as a primary one in UI
                    nucliofunction={"name": self.spec.functions[1]},
                ),
            ]
            if self.spec.canary
            else [
                schemas.APIGatewayUpstream(
                    nucliofunction={"name": function_name},
                )
                for function_name in self.spec.functions
            ]
        )
        if self.spec.ports:
            for i, port in enumerate(self.spec.ports):
                upstreams[i].port = port

        api_gateway = schemas.APIGateway(
            metadata=schemas.APIGatewayMetadata(
                name=self.metadata.name,
                labels=self.metadata.labels,
                annotations=self.metadata.annotations,
            ),
            spec=schemas.APIGatewaySpec(
                name=self.metadata.name,
                description=self.spec.description,
                host=self.spec.host,
                path=self.spec.path,
                authenticationMode=schemas.APIGatewayAuthenticationMode.from_str(
                    self.spec.authentication.authentication_mode
                ),
                upstreams=upstreams,
            ),
            status=schemas.APIGatewayStatus(state=self.status.state),
        )
        api_gateway.spec.authentication = self.spec.authentication.to_scheme()
        return api_gateway

    @property
    def invoke_url(
        self,
    ):
        """
        Get the invoke URL.

        :return: (str) The invoke URL.
        """
        host = self.spec.host
        if not self.spec.host.startswith("http"):
            host = f"https://{self.spec.host}"
        return urljoin(host, self.spec.path).rstrip("/")

    @staticmethod
    def _generate_basic_auth(username: str, password: str):
        token = base64.b64encode(f"{username}:{password}".encode()).decode()
        return f"Basic {token}"

    @staticmethod
    def _resolve_canary(
        upstreams: list[schemas.APIGatewayUpstream],
    ) -> tuple[Union[list[str], None], Union[list[int], None]]:
        if len(upstreams) == 1:
            return [upstreams[0].nucliofunction.get("name")], None
        elif len(upstreams) == 2:
            canary = [0, 0]
            functions = [
                upstreams[0].nucliofunction.get("name"),
                upstreams[1].nucliofunction.get("name"),
            ]
            percentage_1 = upstreams[0].percentage
            percentage_2 = upstreams[1].percentage

            if not percentage_1 and percentage_2:
                percentage_1 = 100 - percentage_2
            if not percentage_2 and percentage_1:
                percentage_2 = 100 - percentage_1
            if percentage_1 and percentage_2:
                canary = [percentage_1, percentage_2]
            return functions, canary
        else:
            # Nuclio only supports 1 or 2 upstream functions
            return None, None

    @property
    def name(self):
        return self.metadata.name

    @name.setter
    def name(self, value):
        self.metadata.name = value

    @property
    def project(self):
        return self.spec.project

    @project.setter
    def project(self, value):
        self.spec.project = value

    @property
    def description(self):
        return self.spec.description

    @description.setter
    def description(self, value):
        self.spec.description = value

    @property
    def host(self):
        return self.spec.host

    @host.setter
    def host(self, value):
        self.spec.host = value

    @property
    def path(self):
        return self.spec.path

    @path.setter
    def path(self, value):
        self.spec.path = value

    @property
    def authentication(self):
        return self.spec.authentication

    @authentication.setter
    def authentication(self, value):
        self.spec.authentication = value



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from collections.abc import Iterator
from datetime import datetime
from typing import Annotated, Any, Callable, Optional, Union

import pydantic.v1

import mlrun.common.schemas.notification as notification_objects
from mlrun.common.types import StrEnum


class EventEntityKind(StrEnum):
    MODEL_ENDPOINT_RESULT = "model-endpoint-result"
    MODEL_MONITORING_APPLICATION = "model-monitoring-application"
    JOB = "job"


class EventEntities(pydantic.v1.BaseModel):
    kind: EventEntityKind
    project: str
    ids: pydantic.v1.conlist(str, min_items=1, max_items=1)


class EventKind(StrEnum):
    DATA_DRIFT_DETECTED = "data-drift-detected"
    DATA_DRIFT_SUSPECTED = "data-drift-suspected"
    CONCEPT_DRIFT_DETECTED = "concept-drift-detected"
    CONCEPT_DRIFT_SUSPECTED = "concept-drift-suspected"
    MODEL_PERFORMANCE_DETECTED = "model-performance-detected"
    MODEL_PERFORMANCE_SUSPECTED = "model-performance-suspected"
    SYSTEM_PERFORMANCE_DETECTED = "system-performance-detected"
    SYSTEM_PERFORMANCE_SUSPECTED = "system-performance-suspected"
    MM_APP_ANOMALY_DETECTED = "mm-app-anomaly-detected"
    MM_APP_ANOMALY_SUSPECTED = "mm-app-anomaly-suspected"
    MM_APP_FAILED = "mm-app-failed"
    FAILED = "failed"


_event_kind_entity_map = {
    EventKind.DATA_DRIFT_SUSPECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.DATA_DRIFT_DETECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.CONCEPT_DRIFT_DETECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.CONCEPT_DRIFT_SUSPECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.MODEL_PERFORMANCE_DETECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.MODEL_PERFORMANCE_SUSPECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.SYSTEM_PERFORMANCE_DETECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.SYSTEM_PERFORMANCE_SUSPECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.MM_APP_ANOMALY_DETECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.MM_APP_ANOMALY_SUSPECTED: [EventEntityKind.MODEL_ENDPOINT_RESULT],
    EventKind.MM_APP_FAILED: [EventEntityKind.MODEL_MONITORING_APPLICATION],
    EventKind.FAILED: [EventEntityKind.JOB],
}


class Event(pydantic.v1.BaseModel):
    kind: EventKind
    timestamp: Union[str, datetime] = None  # occurrence time
    entity: EventEntities
    value_dict: Optional[dict] = pydantic.v1.Field(default_factory=dict)

    def is_valid(self):
        return self.entity.kind in _event_kind_entity_map[self.kind]


class AlertActiveState(StrEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"


class AlertSeverity(StrEnum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


# what should trigger the alert. must be either event (at least 1), or prometheus query
class AlertTrigger(pydantic.v1.BaseModel):
    events: list[EventKind] = []
    prometheus_alert: str = None

    def __eq__(self, other):
        return (
            self.prometheus_alert == other.prometheus_alert
            and self.events == other.events
        )


class AlertCriteria(pydantic.v1.BaseModel):
    count: Annotated[
        int,
        pydantic.v1.Field(
            description="Number of events to wait until notification is sent"
        ),
    ] = 1
    period: Annotated[
        str,
        pydantic.v1.Field(
            description="Time period during which event occurred. e.g. 1d, 3h, 5m, 15s"
        ),
    ] = None

    def __eq__(self, other):
        return self.count == other.count and self.period == other.period


class ResetPolicy(StrEnum):
    MANUAL = "manual"
    AUTO = "auto"


class AlertNotification(pydantic.v1.BaseModel):
    notification: notification_objects.Notification
    cooldown_period: Annotated[
        str,
        pydantic.v1.Field(
            description="Period during which notifications "
            "will not be sent after initial send. The format of this would be in time."
            " e.g. 1d, 3h, 5m, 15s"
        ),
    ] = None


class AlertConfig(pydantic.v1.BaseModel):
    project: str
    id: int = None
    name: str
    description: Optional[str] = ""
    summary: Annotated[
        str,
        pydantic.v1.Field(
            description=(
                "String to be sent in the notifications generated."
                "e.g. 'Model {{project}}/{{entity}} is drifting.'"
                "Supported variables: project, entity, name"
            )
        ),
    ]
    created: Union[str, datetime] = None
    severity: AlertSeverity
    entities: EventEntities
    trigger: AlertTrigger
    criteria: Optional[AlertCriteria]
    reset_policy: ResetPolicy = ResetPolicy.AUTO
    notifications: pydantic.v1.conlist(AlertNotification, min_items=1)
    state: AlertActiveState = AlertActiveState.INACTIVE
    count: Optional[int] = 0
    updated: datetime = None

    class Config:
        extra = pydantic.v1.Extra.allow

    def get_raw_notifications(self) -> list[notification_objects.Notification]:
        return [
            alert_notification.notification for alert_notification in self.notifications
        ]


class AlertsModes(StrEnum):
    enabled = "enabled"
    disabled = "disabled"


class AlertTemplate(
    pydantic.v1.BaseModel
):  # Template fields that are not shared with created configs
    template_id: int = None
    template_name: str
    template_description: Optional[str] = (
        "String explaining the purpose of this template"
    )

    # A property that identifies templates that were created by the system and cannot be modified/deleted by the user
    system_generated: bool = False

    # AlertConfig fields that are pre-defined
    summary: Optional[str] = (
        "String to be sent in the generated notifications e.g. 'Model {{project}}/{{entity}} is drifting.'"
        "See AlertConfig.summary description"
    )
    severity: AlertSeverity
    trigger: AlertTrigger
    criteria: Optional[AlertCriteria]
    reset_policy: ResetPolicy = ResetPolicy.AUTO

    # This is slightly different than __eq__ as it doesn't compare everything
    def templates_differ(self, other):
        return (
            self.template_description != other.template_description
            or self.summary != other.summary
            or self.severity != other.severity
            or self.trigger != other.trigger
            or self.reset_policy != other.reset_policy
            or self.criteria != other.criteria
        )


class AlertActivation(pydantic.v1.BaseModel):
    id: int
    name: str
    project: str
    severity: AlertSeverity
    activation_time: datetime
    entity_id: str
    entity_kind: EventEntityKind
    criteria: AlertCriteria
    event_kind: EventKind
    number_of_events: int
    notifications: list[notification_objects.NotificationState]
    reset_time: Optional[datetime] = None

    def group_key(self, attributes: list[str]) -> Union[Any, tuple]:
        """
        Dynamically create a key for grouping based on the provided attributes.
        - If there's only one attribute, return the value directly (not a single-element tuple).
        - If there are multiple attributes, return them as a tuple for grouping.

        This ensures grouping behaves intuitively without redundant tuple representations.
        """
        if len(attributes) == 1:
            # Avoid single-element tuple like (high,) when only one grouping attribute is used
            return getattr(self, attributes[0])
        # Otherwise, return a tuple of all specified attributes
        return tuple(getattr(self, attr) for attr in attributes)


class AlertActivations(pydantic.v1.BaseModel):
    activations: list[AlertActivation]
    pagination: Optional[dict]

    def __iter__(self) -> Iterator[AlertActivation]:
        return iter(self.activations)

    def __getitem__(self, index: int) -> AlertActivation:
        return self.activations[index]

    def __len__(self) -> int:
        return len(self.activations)

    def group_by(self, *attributes: str) -> dict:
        """
        Group alert activations by specified attributes.

        Args:
        :param attributes: Attributes to group by.

        :returns: A dictionary where keys are tuples of attribute values and values are lists of
            AlertActivation objects.

        Example:
            # Group by project and severity
            grouped = activations.group_by("project", "severity")
        """
        grouped = defaultdict(list)
        for activation in self.activations:
            key = activation.group_key(attributes)
            grouped[key].append(activation)
        return dict(grouped)

    def aggregate_by(
        self,
        group_by_attrs: list[str],
        aggregation_function: Callable[[list[AlertActivation]], Any],
    ) -> dict:
        """
        Aggregate alert activations by specified attributes using a given aggregation function.

        Args:
        :param group_by_attrs: Attributes to group by.
        :param aggregation_function: Function to aggregate grouped activations.

        :returns: A dictionary where keys are tuples of attribute values and values are the result
            of the aggregation function.

        Example:
            # Aggregate by name and entity_id and count number of activations in each group
            activations.aggregate_by(["name", "entity_id"], lambda activations: len(activations))
        """
        grouped = self.group_by(*group_by_attrs)
        aggregated = {
            key: aggregation_function(activations)
            for key, activations in grouped.items()
        }
        return aggregated



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional, Union

import onnx
import onnxoptimizer

import mlrun
from mlrun.artifacts import Artifact

from .._common import ModelHandler


class ONNXModelHandler(ModelHandler):
    """
    Class for handling an ONNX model, enabling loading and saving it during runs.
    """

    # Framework name:
    FRAMEWORK_NAME = "onnx"

    def __init__(
        self,
        model: onnx.ModelProto = None,
        model_path: Optional[str] = None,
        model_name: Optional[str] = None,
        context: mlrun.MLClientCtx = None,
        **kwargs,
    ):
        """
        Initialize the handler. The model can be set here so it won't require loading. Notice that if the model path
        given is of a previously logged model (store model object path), all of the other configurations will be loaded
        automatically as they were logged with the model, hence they are optional.

        :param model:      Model to handle or None in case a loading parameters were supplied.
        :param model_path: Path to the model's directory to load it from. The onnx file must start with the given model
                           name and the directory must contain the onnx file. The model path can be also passed as a
                           model object path in the following format:
                           'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'.
        :param model_name: The model name for saving and logging the model:
                           * Mandatory for loading the model from a local path.
                           * If given a logged model (store model path) it will be read from the artifact.
                           * If given a loaded model object and the model name is None, the name will be set to the
                             model's object name / class.
        :param context:    MLRun context to work with for logging the model.

        :raise MLRunInvalidArgumentError: There was no model or model directory supplied.
        """
        # Setup the base handler class:
        super().__init__(
            model=model,
            model_path=model_path,
            model_name=model_name,
            context=context,
            **kwargs,
        )

    # TODO: output_path won't work well with logging artifacts. Need to look into changing the logic of 'log_artifact'.
    def save(
        self, output_path: Optional[str] = None, **kwargs
    ) -> Union[dict[str, Artifact], None]:
        """
        Save the handled model at the given output path. If a MLRun context is available, the saved model files will be
        logged and returned as artifacts.

        :param output_path: The full path to the directory to save the handled model at. If not given, the context
                            stored will be used to save the model in the default artifacts location.

        :return The saved model additional artifacts (if needed) dictionary if context is available and None otherwise.
        """
        super().save(output_path=output_path)

        # Set the output path:
        if output_path is None:
            output_path = os.path.join(self._context.artifact_path, self._model_name)

        # Save the model:
        self._model_file = f"{self._model_name}.onnx"
        onnx.save(self._model, self._model_file)

        return None

    def load(self, **kwargs):
        """
        Load the specified model in this handler.
        """
        super().load()

        # Check that the model is well-formed:
        # TODO: Currently not working well with HuggingFace models so we skip it
        # onnx.checker.check_model(self._model_file)

        # Load the ONNX model:
        self._model = onnx.load(self._model_file)

    def optimize(
        self, optimizations: Optional[list[str]] = None, fixed_point: bool = False
    ):
        """
        Use ONNX optimizer to optimize the ONNX model. The optimizations supported can be seen by calling
        'onnxoptimizer.get_available_passes()'

        :param optimizations: List of possible optimizations. If None, all of the optimizations will be used. Default:
                              None.
        :param fixed_point:   Optimize the weights using fixed point. Default: False.
        """
        # Set the ONNX optimizations list:
        onnx_optimizations = onnxoptimizer.get_fuse_and_elimination_passes()
        if optimizations is None:
            # Set to all optimizations:
            optimizations = onnx_optimizations

        # Optimize the model:
        self._model = onnxoptimizer.optimize(
            self._model, passes=optimizations, fixed_point=fixed_point
        )

    def to_onnx(self, *args, **kwargs) -> onnx.ModelProto:
        """
        Convert the model in this handler to an ONNX model. In this case the handled ONNX model will simply be returned.

        :return: The current handled ONNX model as there is nothing to convert.
        """
        return self._model

    def _collect_files_from_local_path(self):
        """
        If the model path given is of a local path, search for the needed model files and collect them into this handler
        for later loading the model.

        :raise MLRunNotFoundError: If the onnx file was not found.
        """
        self._model_file = os.path.join(self._model_path, f"{self._model_name}.onnx")
        if not os.path.exists(self._model_file):
            raise mlrun.errors.MLRunNotFoundError(
                f"The model file '{self._model_name}.onnx' was not found within the given "
                f"'model_path': '{self._model_path}'"
            )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from functools import partial
from typing import Optional

from mergedeep import merge

import mlrun
import mlrun.errors
from mlrun.artifacts.llm_prompt import LLMPromptArtifact
from mlrun.artifacts.model import ModelArtifact
from mlrun.datastore.datastore_profile import datastore_profile_read
from mlrun.datastore.model_provider.model_provider import (
    ModelProvider,
)
from mlrun.datastore.remote_client import BaseRemoteClient
from mlrun.datastore.utils import (
    parse_url,
)
from mlrun.errors import err_to_str
from mlrun.utils.helpers import get_local_file_schema

from ..artifacts.base import verify_target_path
from ..utils import DB_SCHEMA, RunKeys
from .base import DataItem, DataStore, HttpStore
from .filestore import FileStore
from .inmem import InMemoryStore
from .model_provider.huggingface_provider import HuggingFaceProvider
from .model_provider.mock_model_provider import MockModelProvider
from .model_provider.openai_provider import OpenAIProvider
from .store_resources import get_store_resource, is_store_uri
from .v3io import V3ioStore

in_memory_store = InMemoryStore()


def schema_to_store(schema) -> DataStore.__subclasses__():
    # import store classes inside to enable making their dependencies optional (package extras)

    if not schema or schema in get_local_file_schema():
        return FileStore
    elif schema == "s3":
        try:
            from .s3 import S3Store
        except ImportError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "s3 packages are missing, use pip install mlrun[s3]"
            )

        return S3Store
    elif schema in ["az", "wasbs", "wasb"]:
        try:
            from .azure_blob import AzureBlobStore
        except ImportError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "azure blob storage packages are missing, use pip install mlrun[azure-blob-storage]"
            )

        return AzureBlobStore
    elif schema in ["v3io", "v3ios"]:
        return V3ioStore
    elif schema in ["redis", "rediss"]:
        from .redis import RedisStore

        return RedisStore
    elif schema in ["http", "https"]:
        return HttpStore
    elif schema in ["gcs", "gs"]:
        try:
            from .google_cloud_storage import GoogleCloudStorageStore
        except ImportError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "Google cloud storage packages are missing, use pip install mlrun[google-cloud-storage]"
            )
        return GoogleCloudStorageStore
    elif schema == "dbfs":
        from .dbfs_store import DBFSStore

        return DBFSStore
    elif schema in ["hdfs", "webhdfs"]:
        from .hdfs import HdfsStore

        return HdfsStore
    elif schema == "oss":
        from .alibaba_oss import OSSStore

        return OSSStore
    raise ValueError(f"unsupported store scheme ({schema})")


def schema_to_model_provider(
    schema: str, raise_missing_schema_exception=True
) -> type[ModelProvider]:
    schema_dict = {
        "openai": OpenAIProvider,
        "huggingface": HuggingFaceProvider,
        "mock": MockModelProvider,
    }
    provider_class = schema_dict.get(schema, None)
    if not provider_class:
        if raise_missing_schema_exception:
            raise ValueError(f"unsupported model provider schema ({schema})")
        else:
            warnings.warn(f"unsupported model provider schema: {schema}")
    return provider_class


def uri_to_ipython(link):
    schema, endpoint, parsed_url = parse_url(link)
    if schema in [DB_SCHEMA, "memory", "ds"]:
        return ""
    return schema_to_store(schema).uri_to_ipython(endpoint, parsed_url.path)


class StoreManager:
    def __init__(self, secrets=None, db=None):
        self._stores = {}
        self._secrets = secrets or {}
        self._db = db

    def set(self, secrets=None, db=None):
        if db and not self._db:
            self._db = db
        if secrets:
            for key, val in secrets.items():
                self._secrets[key] = val
        return self

    def _get_db(self):
        if not self._db:
            self._db = mlrun.get_run_db(secrets=self._secrets)
        return self._db

    def from_dict(self, struct: dict):
        stor_list = struct.get(RunKeys.data_stores)
        if stor_list and isinstance(stor_list, list):
            for stor in stor_list:
                schema, endpoint, parsed_url = parse_url(stor.get("url"))
                new_stor = schema_to_store(schema)(self, schema, stor["name"], endpoint)
                new_stor.subpath = parsed_url.path
                new_stor.secret_pfx = stor.get("secret_pfx")
                new_stor.options = stor.get("options", {})
                new_stor.from_spec = True
                self._stores[stor["name"]] = new_stor

    def to_dict(self, struct):
        struct[RunKeys.data_stores] = [
            stor.to_dict() for stor in self._stores.values() if stor.from_spec
        ]

    def secret(self, key):
        return self._secrets.get(key)

    def _add_store(self, store):
        self._stores[store.name] = store

    def get_store_artifact(
        self,
        url,
        project="",
        allow_empty_resources=None,
        secrets=None,
    ):
        """
        This is expected to be run only on client side. server is not expected to load artifacts.
        """
        try:
            resource = get_store_resource(
                url,
                db=self._get_db(),
                secrets=self._secrets,
                project=project,
                data_store_secrets=secrets,
            )
        except Exception as exc:
            raise OSError(f"artifact {url} not found, {err_to_str(exc)}")
        target = resource.get_target_path()

        # the allow_empty.. flag allows us to have functions which dont depend on having targets e.g. a function
        # which accepts a feature vector uri and generate the offline vector (parquet) for it if it doesnt exist
        if not allow_empty_resources:
            if isinstance(resource, LLMPromptArtifact):
                if not resource.spec.model_uri:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"LLMPromptArtifact {url} does not contain model artifact uri"
                    )
            elif not target and not (
                isinstance(resource, ModelArtifact) and resource.model_url
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Resource {url} does not have a valid/persistent offline target or model_url"
                )
        return resource, target or ""

    def object(
        self,
        url,
        key="",
        project="",
        allow_empty_resources=None,
        secrets: Optional[dict] = None,
        **kwargs,
    ) -> DataItem:
        meta = artifact_url = None
        if is_store_uri(url):
            artifact_url = url
            meta, url = self.get_store_artifact(
                url, project, allow_empty_resources, secrets
            )
            if not allow_empty_resources:
                verify_target_path(meta)

        store, subpath, url = self.get_or_create_store(
            url, secrets=secrets, project_name=project
        )
        return DataItem(
            key,
            store,
            subpath,
            url,
            meta=meta,
            artifact_url=artifact_url,
        )

    def _get_or_create_remote_client(
        self,
        url,
        secrets: Optional[dict] = None,
        project_name="",
        cache: Optional[dict] = None,
        schema_to_class: callable = schema_to_store,
        **kwargs,
    ) -> (BaseRemoteClient, str, str):
        # The cache can be an empty dictionary ({}), even if it is a _stores object
        cache = cache if cache is not None else {}
        schema, endpoint, parsed_url = parse_url(url)
        subpath = parsed_url.path
        cache_key = f"{schema}://{endpoint}" if endpoint else f"{schema}://"

        if schema == "ds":
            datastore_profile = datastore_profile_read(url, project_name, secrets)
            secrets = merge({}, secrets or {}, datastore_profile.secrets() or {})
            url = datastore_profile.url(subpath)
            schema, endpoint, parsed_url = parse_url(url)
            subpath = parsed_url.path

        if schema == "memory":
            subpath = url[len("memory://") :]
            return in_memory_store, subpath, url

        elif schema in get_local_file_schema():
            # parse_url() will drop the windows drive-letter from the path for url like "c:\a\b".
            # As a workaround, we set subpath to the url.
            subpath = url.replace("file://", "", 1)

        if not schema and endpoint:
            if endpoint in cache.keys():
                return cache[endpoint], subpath, url
            else:
                raise ValueError(f"no such store ({endpoint})")

        if not secrets and not mlrun.config.is_running_as_api():
            if cache_key in cache.keys():
                return cache[cache_key], subpath, url

        # support u/p embedding in url (as done in redis) by setting netloc as the "endpoint" parameter
        # when running on server we don't cache the datastore, because there are multiple users and we don't want to
        # cache the credentials, so for each new request we create a new store
        remote_client_class = schema_to_class(schema)
        remote_client = None
        if remote_client_class:
            endpoint, subpath = remote_client_class.parse_endpoint_and_path(
                endpoint, subpath
            )
            remote_client = remote_client_class(
                self, schema, cache_key, endpoint, secrets=secrets, **kwargs
            )
            if not secrets and not mlrun.config.is_running_as_api():
                cache[cache_key] = remote_client
        else:
            warnings.warn("scheme not found. Returning None")
        return remote_client, subpath, url

    def get_or_create_store(
        self,
        url,
        secrets: Optional[dict] = None,
        project_name="",
    ) -> (DataStore, str, str):
        datastore, sub_path, url = self._get_or_create_remote_client(
            url=url,
            secrets=secrets,
            project_name=project_name,
            cache=self._stores,
            schema_to_class=schema_to_store,
        )
        if not isinstance(datastore, DataStore):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "remote client by url is not datastore"
            )
        return datastore, sub_path, url

    def get_or_create_model_provider(
        self,
        url,
        secrets: Optional[dict] = None,
        project_name="",
        default_invoke_kwargs: Optional[dict] = None,
        raise_missing_schema_exception=True,
    ) -> ModelProvider:
        schema_to_provider_with_raise = partial(
            schema_to_model_provider,
            raise_missing_schema_exception=raise_missing_schema_exception,
        )
        model_provider, _, _ = self._get_or_create_remote_client(
            url=url,
            secrets=secrets,
            project_name=project_name,
            schema_to_class=schema_to_provider_with_raise,
            default_invoke_kwargs=default_invoke_kwargs,
        )
        if model_provider and not isinstance(model_provider, ModelProvider):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "remote client by url is not model_provider"
            )
        return model_provider

    def reset_secrets(self):
        self._secrets = {}

    def model_provider_object(
        self,
        url,
        project="",
        allow_empty_resources=None,
        secrets: Optional[dict] = None,
        default_invoke_kwargs: Optional[dict] = None,
        raise_missing_schema_exception=True,
    ) -> ModelProvider:
        if mlrun.datastore.is_store_uri(url):
            resource = self.get_store_artifact(
                url,
                project,
                allow_empty_resources,
                secrets,
            )
            if not isinstance(resource, ModelArtifact) or not resource.model_url:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "unable to create the model provider from the given resource URI"
                )
            url = resource.model_url
            default_invoke_kwargs = default_invoke_kwargs or resource.default_config
        model_provider = self.get_or_create_model_provider(
            url,
            secrets=secrets,
            project_name=project,
            default_invoke_kwargs=default_invoke_kwargs,
            raise_missing_schema_exception=raise_missing_schema_exception,
        )
        return model_provider



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
from pathlib import Path
from typing import Optional

from fsspec.registry import get_filesystem_class
from google.auth.credentials import Credentials
from google.cloud.storage import Client, transfer_manager
from google.oauth2 import service_account

import mlrun.errors
from mlrun.utils import logger

from .base import DataStore, FileStats, make_datastore_schema_sanitizer

# Google storage objects will be represented with the following URL: gcs://<bucket name>/<path> or gs://...


class GoogleCloudStorageStore(DataStore):
    using_bucket = True
    workers = 8
    chunk_size = 32 * 1024 * 1024

    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        super().__init__(parent, name, schema, endpoint, secrets=secrets)
        self._storage_client = None
        self._storage_options = None

    @property
    def storage_client(self):
        if self._storage_client:
            return self._storage_client

        token = self._get_credentials().get("token")
        access = "https://www.googleapis.com/auth/devstorage.full_control"
        if isinstance(token, str):
            if os.path.exists(token):
                credentials = service_account.Credentials.from_service_account_file(
                    token, scopes=[access]
                )
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "gcsfs authentication file not found!"
                )
        elif isinstance(token, dict):
            credentials = service_account.Credentials.from_service_account_info(
                token, scopes=[access]
            )
        elif isinstance(token, Credentials):
            credentials = token
        else:
            raise ValueError(f"Unsupported token type: {type(token)}")
        self._storage_client = Client(credentials=credentials)
        return self._storage_client

    @property
    def filesystem(self):
        """return fsspec file system object, if supported"""
        if not self._filesystem:
            filesystem_class = get_filesystem_class(protocol=self.kind)
            self._filesystem = make_datastore_schema_sanitizer(
                filesystem_class,
                using_bucket=self.using_bucket,
                **self.storage_options,
            )
        return self._filesystem

    @property
    def storage_options(self):
        if self._storage_options:
            return self._storage_options
        credentials = self._get_credentials()
        # due to caching problem introduced in gcsfs 2024.3.1 (ML-7636)
        credentials["use_listings_cache"] = False
        self._storage_options = credentials
        return self._storage_options

    def _get_credentials(self):
        credentials = self._get_secret_or_env(
            "GCP_CREDENTIALS"
        ) or self._get_secret_or_env("GOOGLE_APPLICATION_CREDENTIALS")
        if credentials:
            try:
                # Try to handle credentials as a json connection string or do nothing if already a dict
                token = (
                    credentials
                    if isinstance(credentials, dict)
                    else json.loads(credentials)
                )
            except json.JSONDecodeError:
                # If it's not json, handle it as a filename
                token = credentials
            return self._sanitize_options(dict(token=token))
        else:
            logger.info(
                "No GCS credentials available - auth will rely on auto-discovery of credentials"
            )
            return self._sanitize_options(None)

    def get_storage_options(self):
        return self.storage_options

    def _make_path(self, key):
        key = key.strip("/")
        path = Path(self.endpoint, key).as_posix()
        return path

    def get(self, key, size=None, offset=0):
        path = self._make_path(key)

        end = offset + size if size else None
        blob = self.filesystem.cat_file(path, start=offset, end=end)
        return blob

    def put(self, key, data, append=False):
        path = self._make_path(key)

        if append:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Append mode not supported for Google cloud storage datastore"
            )
        data, mode = self._prepare_put_data(data, append)
        with self.filesystem.open(path, mode) as f:
            f.write(data)

    def upload(self, key, src_path):
        file_size = os.path.getsize(src_path)
        united_path = self._make_path(key)

        # Multiple upload limitation recommendations as described in
        # https://cloud.google.com/storage/docs/multipart-uploads#storage-upload-object-chunks-python

        if file_size <= self.chunk_size:
            self.filesystem.put_file(src_path, united_path, overwrite=True)
            return

        bucket = self.storage_client.bucket(self.endpoint)
        blob = bucket.blob(key.strip("/"))

        try:
            transfer_manager.upload_chunks_concurrently(
                src_path, blob, chunk_size=self.chunk_size, max_workers=self.workers
            )
        except Exception as upload_chunks_concurrently_exception:
            logger.warning(
                f"gcs: failed to concurrently upload {src_path},"
                f" exception: {upload_chunks_concurrently_exception}. Retrying with single part upload."
            )
            self.filesystem.put_file(src_path, united_path, overwrite=True)

    def stat(self, key):
        path = self._make_path(key)

        files = self.filesystem.ls(path, detail=True)
        if len(files) == 1 and files[0]["type"] == "file":
            size = files[0]["size"]
            modified = files[0]["updated"]
        elif len(files) == 1 and files[0]["type"] == "directory":
            raise FileNotFoundError("Operation expects a file not a directory!")
        else:
            raise ValueError("Operation expects to receive a single file!")
        return FileStats(size, modified)

    def listdir(self, key):
        path = self._make_path(key)
        if self.filesystem.isfile(path):
            return key
        remote_path = f"{path}/**"
        files = self.filesystem.glob(remote_path)
        key_length = len(key)
        files = [
            f.split("/", 1)[1][key_length:] for f in files if len(f.split("/")) > 1
        ]
        return files

    def rm(self, path, recursive=False, maxdepth=None):
        path = self._make_path(path)
        # in order to raise an error in case of a connection error (ML-7056)
        self.filesystem.exists(path)
        super().rm(path, recursive=recursive, maxdepth=maxdepth)

    def get_spark_options(self, path=None):
        res = {}
        st = self._get_credentials()
        if "token" in st:
            res = {"spark.hadoop.google.cloud.auth.service.account.enable": "true"}
            if isinstance(st["token"], str):
                # Token is a filename, read json from it
                with open(st["token"]) as file:
                    credentials = json.load(file)
            else:
                # Token is a dictionary, use it directly
                credentials = st["token"]

            if "project_id" in credentials:
                res["spark.hadoop.fs.gs.project.id"] = credentials["project_id"]
            if "private_key_id" in credentials:
                res["spark.hadoop.fs.gs.auth.service.account.private.key.id"] = (
                    credentials["private_key_id"]
                )
            if "private_key" in credentials:
                res["spark.hadoop.fs.gs.auth.service.account.private.key"] = (
                    credentials["private_key"]
                )
            if "client_email" in credentials:
                res["spark.hadoop.fs.gs.auth.service.account.email"] = credentials[
                    "client_email"
                ]
            if "client_id" in credentials:
                res["spark.hadoop.fs.gs.client.id"] = credentials["client_id"]
        return res

    @property
    def spark_url(self):
        return f"gs://{self.endpoint}"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import builtins
import http
import importlib.util as imputil
import os
import tempfile
import typing
import uuid

import mlrun
import mlrun.common.constants as mlrun_constants
import mlrun.common.runtimes.constants
import mlrun.common.schemas
import mlrun.common.schemas.function
import mlrun.common.schemas.workflow
import mlrun.utils.notifications
import mlrun_pipelines.common.models
import mlrun_pipelines.patcher
import mlrun_pipelines.utils
from mlrun.errors import err_to_str
from mlrun.utils import (
    get_workflow_url,
    logger,
    normalize_workflow_name,
    retry_until_successful,
)

from ..common.helpers import parse_versioned_object_uri
from ..config import config
from ..run import (
    _run_pipeline,
    retry_pipeline,
    terminate_pipeline,
    wait_for_pipeline_completion,
)
from ..runtimes.pod import AutoMountType


def get_workflow_engine(engine_kind, local=False):
    if pipeline_context.is_run_local(local):
        if engine_kind == mlrun.common.schemas.workflow.EngineType.KFP:
            logger.warning(
                "Running kubeflow pipeline locally, note some ops may not run locally!"
            )
        elif engine_kind == mlrun.common.schemas.workflow.EngineType.REMOTE:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Cannot run a remote pipeline locally using `kind='remote'` and `local=True`. "
                "in order to run a local pipeline remotely, please use `engine='remote:local'` instead"
            )
        return _LocalRunner
    if not engine_kind or engine_kind == mlrun.common.schemas.workflow.EngineType.KFP:
        return _KFPRunner
    if engine_kind == mlrun.common.schemas.workflow.EngineType.LOCAL:
        return _LocalRunner
    if engine_kind == mlrun.common.schemas.workflow.EngineType.REMOTE:
        return _RemoteRunner
    raise mlrun.errors.MLRunInvalidArgumentError(
        f"Provided workflow engine is not supported. engine_kind={engine_kind}"
    )


class WorkflowSpec(mlrun.model.ModelObj):
    """workflow spec and helpers"""

    def __init__(
        self,
        engine: typing.Optional[str] = None,
        code: typing.Optional[str] = None,
        path: typing.Optional[str] = None,
        args: typing.Optional[dict] = None,
        name: typing.Optional[str] = None,
        handler: typing.Optional[str] = None,
        args_schema: typing.Optional[dict] = None,
        schedule: typing.Union[str, mlrun.common.schemas.ScheduleCronTrigger] = None,
        cleanup_ttl: typing.Optional[int] = None,
        image: typing.Optional[str] = None,
        workflow_runner_node_selector: typing.Optional[dict[str, str]] = None,
    ):
        self.engine = engine
        self.code = code
        self.path = path
        self.args = args
        self.name = name
        self.handler = handler
        self.cleanup_ttl = cleanup_ttl
        self.args_schema = args_schema
        self.run_local = False
        self._tmp_path = None
        self.schedule = schedule
        self.image = image
        self.workflow_runner_node_selector = workflow_runner_node_selector

    def get_source_file(self, context=""):
        if not self.code and not self.path:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "workflow must have code or path properties"
            )
        if self.code:
            with tempfile.NamedTemporaryFile(
                mode="w", suffix=".py", delete=False
            ) as workflow_fh:
                workflow_fh.write(self.code)
                self._tmp_path = workflow_path = workflow_fh.name
        else:
            workflow_path = self.path or ""
            if (
                context
                and not workflow_path.startswith("/")
                # since the user may provide a path the includes the context,
                # we need to make sure we don't add it twice
                and not workflow_path.startswith(context)
            ):
                workflow_path = os.path.join(context, workflow_path.lstrip("./"))
        return workflow_path

    def merge_args(self, extra_args):
        self.args = self.args or {}
        required = []
        if self.args_schema:
            for schema in self.args_schema:
                name = schema.get("name")
                if name not in self.args:
                    self.args[name] = schema.get("default")
                if schema.get("required"):
                    required.append(name)
        if extra_args:
            for k, v in extra_args.items():
                self.args[k] = v
                if k in required:
                    required.remove(k)
        if required:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Workflow argument(s) {','.join(required)} are required and were not specified"
            )

    def clear_tmp(self):
        if self._tmp_path:
            os.remove(self._tmp_path)


class FunctionsDict:
    """Virtual dictionary hosting the project functions, cached or in the DB"""

    def __init__(self, project, decorator=None):
        self.project = project
        self._decorator = decorator

    @property
    def _functions(self):
        return self.project.spec._function_objects

    def enrich(self, function, key):
        enriched_function = enrich_function_object(
            self.project, function, self._decorator
        )
        self._functions[key] = enriched_function  # update the cache
        return self._functions[key]

    def load_or_set_function(self, key, default=None) -> mlrun.runtimes.BaseRuntime:
        try:
            function = self.project.get_function(key, sync=False)
        except Exception as e:
            if not default:
                raise e
            function = default

        return self.enrich(function, key)

    def get(self, key, default=None) -> mlrun.runtimes.BaseRuntime:
        return self.load_or_set_function(key, default)

    def __getitem__(self, key) -> mlrun.runtimes.BaseRuntime:
        return self.load_or_set_function(key)

    def __setitem__(self, key, val):
        self._functions[key] = val

    def values(self):
        return [self.enrich(function, key) for key, function in self._functions.items()]

    def keys(self):
        return self._functions.keys()

    def items(self):
        return [
            (key, self.enrich(function, key))
            for key, function in self._functions.items()
        ]

    def __len__(self):
        return len(self._functions)

    def __iter__(self):
        yield from self._functions.keys()

    def __delitem__(self, key):
        del self._functions[key]


class _PipelineContext:
    """current (running) pipeline context"""

    def __init__(self):
        self.project = None
        self.workflow = None
        self.functions = FunctionsDict(None)
        self.workflow_id = None
        self.workflow_artifact_path = None
        self.runs_map = {}

    def is_run_local(self, local=None):
        if local is not None:
            # if the user specified an explicit value in local we use it
            return local
        force_run_local = mlrun.mlconf.force_run_local
        if force_run_local is None or force_run_local == "auto":
            force_run_local = not mlrun.mlconf.is_api_running_on_k8s()

        if self.workflow:
            if not mlrun.mlconf.kfp_url:
                logger.debug("Kubeflow pipeline URL is not set, running locally")
                force_run_local = True
            force_run_local = force_run_local or self.workflow.run_local

        return force_run_local

    def set(self, project, workflow=None):
        self.project = project
        self.workflow = workflow
        self.functions.project = project
        self.runs_map = {}

    def clear(self, with_project=False):
        if with_project:
            self.project = None
            self.functions.project = None
        self.workflow = None
        self.runs_map = {}
        self.workflow_id = None
        self.workflow_artifact_path = None

    def is_initialized(self, raise_exception=False):
        if self.project:
            return True
        if raise_exception:
            raise ValueError(
                "Pipeline context is not initialized, must be used inside a pipeline"
            )
        return False


pipeline_context = _PipelineContext()


def _set_function_attribute_on_kfp_pod(
    kfp_pod_template, function, pod_template_key, function_spec_key
):
    try:
        if kfp_pod_template.get("name").startswith(function.metadata.name):
            attribute_value = getattr(function.spec, function_spec_key, None)
            if attribute_value:
                kfp_pod_template[pod_template_key] = attribute_value
    except Exception as err:
        kfp_pod_name = kfp_pod_template.get("name")
        logger.warning(
            f"Unable to set function attribute on kfp pod {kfp_pod_name}",
            function_spec_key=function_spec_key,
            pod_template_key=pod_template_key,
            error=err_to_str(err),
        )


def _enrich_kfp_pod_security_context(kfp_pod_template, function):
    if (
        mlrun.runtimes.RuntimeKinds.is_local_runtime(function.kind)
        or mlrun.mlconf.function.spec.security_context.enrichment_mode
        == mlrun.common.schemas.SecurityContextEnrichmentModes.disabled.value
    ):
        return

    # ensure kfp pod user id is not None or 0 (root)
    if not mlrun.mlconf.function.spec.security_context.pipelines.kfp_pod_user_unix_id:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"Kubeflow pipeline pod user id is invalid: "
            f"{mlrun.mlconf.function.spec.security_context.pipelines.kfp_pod_user_unix_id}, "
            f"it must be an integer greater than 0. "
            f"See mlrun.config.function.spec.security_context.pipelines.kfp_pod_user_unix_id for more details."
        )

    kfp_pod_user_unix_id = int(
        mlrun.mlconf.function.spec.security_context.pipelines.kfp_pod_user_unix_id
    )
    kfp_pod_template["SecurityContext"] = {
        "runAsUser": kfp_pod_user_unix_id,
        "runAsGroup": mlrun.mlconf.get_security_context_enrichment_group_id(
            kfp_pod_user_unix_id
        ),
    }


def get_db_function(project, key) -> mlrun.runtimes.BaseRuntime:
    project_instance, name, tag, hash_key = parse_versioned_object_uri(
        key, project.metadata.name
    )
    runtime = mlrun.get_run_db().get_function(name, project_instance, tag, hash_key)
    return mlrun.new_function(runtime=runtime)


def enrich_function_object(
    project: mlrun.common.schemas.Project,
    function: mlrun.runtimes.BaseRuntime,
    decorator: typing.Optional[typing.Callable] = None,
    copy_function: bool = True,
    try_auto_mount: bool = True,
) -> mlrun.runtimes.BaseRuntime:
    if hasattr(function, "_enriched"):
        return function
    f = function.copy() if copy_function else function
    f.metadata.project = project.metadata.name
    setattr(f, "_enriched", True)

    # set project default image if defined and function does not have an image specified
    if project.spec.default_image and not f.spec.image:
        f._enriched_image = True
        f.spec.image = project.spec.default_image

    src = f.spec.build.source
    if src and src in [".", "./"]:
        if not project.spec.source and not project.spec.mountdir:
            logger.warning(
                "project.spec.source should be specified when function is using code from project context"
            )

        if project.spec.mountdir:
            f.spec.workdir = project.spec.mountdir
            f.spec.build.source = ""
        else:
            f.spec.build.source = project.spec.source
            f.spec.build.load_source_on_run = project.spec.load_source_on_run
            f.spec.build.source_code_target_dir = (
                project.spec.build.source_code_target_dir
            )
            f.spec.workdir = project.spec.workdir or project.spec.subpath
            f.prepare_image_for_deploy()

    if project.spec.default_requirements:
        f.with_requirements(project.spec.default_requirements)
    if decorator:
        decorator(f)

    if project.spec.default_function_node_selector:
        f.enrich_runtime_spec(
            project.spec.default_function_node_selector,
        )
    if try_auto_mount:
        if (
            decorator and AutoMountType.is_auto_modifier(decorator)
        ) or project.spec.disable_auto_mount:
            f.spec.disable_auto_mount = True
        f.try_auto_mount_based_on_config()

    return f


class _PipelineRunStatus:
    """pipeline run result (status)"""

    def __init__(
        self,
        run_id: str,
        engine: type["_PipelineRunner"],
        project: "mlrun.projects.MlrunProject",
        workflow: WorkflowSpec = None,
        state: mlrun_pipelines.common.models.RunStatuses = "",
        exc: typing.Optional[Exception] = None,
    ):
        """
        :param run_id:      unique id of the pipeline run
        :param engine:      pipeline runner
        :param project:     mlrun project
        :param workflow:    workflow with spec on how to run the pipeline
        :param state:       the current state of the pipeline run
        :param exc:         exception that was raised during the pipeline run
        """
        self.run_id = run_id
        self.project = project
        self.workflow = workflow
        self._engine = engine
        self._state = state
        self._exc = exc

    @property
    def state(self):
        if (
            self._state
            not in mlrun_pipelines.common.models.RunStatuses.stable_statuses()
        ):
            self._state = self._engine.get_state(self.run_id, self.project)
        return self._state

    @property
    def exc(self):
        return self._exc

    def wait_for_completion(self, timeout=None, expected_statuses=None):
        returned_state = self._engine.wait_for_completion(
            self,
            project=self.project,
            timeout=timeout,
            expected_statuses=expected_statuses,
        )
        # TODO: returning a state is optional until all runners implement wait_for_completion
        if returned_state:
            self._state = returned_state
        return self._state

    def retry(self) -> str:
        run_id = self._engine.retry(
            self,
            project=self.project,
        )
        return run_id

    def __str__(self):
        return str(self.run_id)

    def __repr__(self):
        return str(self.run_id)


class _PipelineRunner(abc.ABC):
    """abstract pipeline runner class"""

    engine = ""

    @classmethod
    @abc.abstractmethod
    def save(cls, project, workflow_spec: WorkflowSpec, target, artifact_path=None):
        raise NotImplementedError(
            f"Save operation not supported in {cls.engine} pipeline engine"
        )

    @classmethod
    @abc.abstractmethod
    def retry(
        cls,
        run: "_PipelineRunStatus",
        project: typing.Optional["mlrun.projects.MlrunProject"] = None,
    ) -> str:
        raise NotImplementedError(
            f"Retry operation not supported in {cls.engine} pipeline engine"
        )

    @classmethod
    @abc.abstractmethod
    def run(
        cls,
        project,
        workflow_spec: WorkflowSpec,
        name=None,
        workflow_handler=None,
        secrets=None,
        artifact_path=None,
        namespace=None,
        source=None,
        notifications: typing.Optional[list[mlrun.model.Notification]] = None,
        context: typing.Optional[mlrun.execution.MLClientCtx] = None,
    ) -> _PipelineRunStatus:
        pass

    @staticmethod
    @abc.abstractmethod
    def wait_for_completion(
        run: "_PipelineRunStatus",
        project: typing.Optional["mlrun.projects.MlrunProject"] = None,
        timeout: typing.Optional[int] = None,
        expected_statuses: typing.Optional[list[str]] = None,
    ):
        pass

    @staticmethod
    @abc.abstractmethod
    def get_state(run_id, project=None):
        pass

    @staticmethod
    def get_run_status(
        project,
        run: _PipelineRunStatus,
        timeout=None,
        expected_statuses=None,
        notifiers: mlrun.utils.notifications.CustomNotificationPusher = None,
        **kwargs,
    ):
        timeout = timeout or 60 * 60
        raise_error = None
        state = ""
        try:
            if timeout:
                state = run.wait_for_completion(
                    timeout=timeout, expected_statuses=expected_statuses
                )
        except RuntimeError as exc:
            # push runs table also when we have errors
            raise_error = exc

        mldb = mlrun.db.get_run_db(secrets=project._secrets)
        runs = mldb.list_runs(project=project.name, labels=f"workflow={run.run_id}")

        # TODO: The below section duplicates notifiers.push_pipeline_run_results() logic. We should use it instead.
        errors_counter = 0
        for r in runs:
            if r["status"].get("state", "") == "error":
                errors_counter += 1

        text = _PipelineRunner._generate_workflow_finished_message(
            run.run_id, errors_counter, run._state
        )
        notifiers = notifiers or project.notifiers
        if notifiers:
            notifiers.push(text, "info", runs)

        project.push_pipeline_notification_kfp_runner(run.run_id, run._state, text)

        if raise_error:
            raise raise_error
        return state or run._state, errors_counter, text

    @staticmethod
    def _get_handler(workflow_handler, workflow_spec, project, secrets):
        if not (workflow_handler and callable(workflow_handler)):
            workflow_file = workflow_spec.get_source_file(project.spec.get_code_path())
            workflow_handler = create_pipeline(
                project,
                workflow_file,
                pipeline_context.functions,
                secrets,
                handler=workflow_handler or workflow_spec.handler,
            )
        else:
            builtins.funcs = pipeline_context.functions
        return workflow_handler

    @staticmethod
    def _generate_workflow_finished_message(run_id, errors_counter, state):
        text = f"Workflow {run_id} finished"
        if errors_counter:
            text += f" with {errors_counter} errors"
        if state:
            text += f", state={state}"
        return text


class _KFPRunner(_PipelineRunner):
    """Kubeflow pipelines runner"""

    engine = "kfp"

    @classmethod
    def save(cls, project, workflow_spec: WorkflowSpec, target, artifact_path=None):
        pipeline_context.set(project, workflow_spec)
        workflow_file = workflow_spec.get_source_file(project.spec.get_code_path())
        functions = FunctionsDict(project)
        pipeline = create_pipeline(
            project,
            workflow_file,
            functions,
            secrets=project._secrets,
        )
        mlrun_pipelines.utils.compile_pipeline(
            artifact_path=artifact_path or project.spec.artifact_path,
            cleanup_ttl=workflow_spec.cleanup_ttl,
            ops=None,
            pipeline=pipeline,
            pipe_file=target,
            type_check=True,
        )
        workflow_spec.clear_tmp()
        pipeline_context.clear()

    @classmethod
    def run(
        cls,
        project: "mlrun.projects.MlrunProject",
        workflow_spec: WorkflowSpec,
        name=None,
        workflow_handler=None,
        secrets=None,
        artifact_path=None,
        namespace=None,
        source=None,
        notifications: typing.Optional[list[mlrun.model.Notification]] = None,
        context: typing.Optional[mlrun.execution.MLClientCtx] = None,
    ) -> _PipelineRunStatus:
        pipeline_context.set(project, workflow_spec)
        workflow_handler = _PipelineRunner._get_handler(
            workflow_handler, workflow_spec, project, secrets
        )
        if source:
            project.set_source(source=source)

        namespace = namespace or config.namespace

        # fallback to old notification behavior
        if notifications:
            logger.warning(
                "Setting notifications on kfp pipeline runner uses old notification behavior. "
                "Notifications will only be sent if you wait for pipeline completion. "
                "Some of the features (like setting message or severity level) are not supported."
            )
            for notification in notifications or []:
                params = notification.params
                params.update(notification.secret_params)
                project.notifiers.add_notification(
                    notification_type=notification.kind,
                    params=params,
                    name=notification.name,
                    message=notification.message,
                    severity=notification.severity,
                    when=notification.when,
                    condition=notification.condition,
                    secret_params=notification.secret_params,
                )

        project.spec.notifications = project.notifiers.server_notifications

        run_id = _run_pipeline(
            workflow_handler,
            project=project.metadata.name,
            arguments=workflow_spec.args,
            experiment=name or workflow_spec.name,
            namespace=namespace,
            artifact_path=artifact_path,
            cleanup_ttl=workflow_spec.cleanup_ttl,
            timeout=int(mlrun.mlconf.workflows.timeouts.kfp),
        )

        # The user provided workflow code might have made changes to function specs that require cleanup
        for func in project.spec._function_objects.values():
            try:
                func.spec.discard_changes()
            except AttributeError:
                logger.debug(
                    "Function does not require a field rollback", func_type=type(func)
                )
            except Exception as exc:
                logger.warning(
                    "Failed to rollback spec fields for function",
                    project=project,
                    func_name=func.metadata.name,
                    exc_info=err_to_str(exc),
                )

        # Pushing only relevant notification for the client (ipython and console)
        project.notifiers.push_pipeline_start_message_from_client(
            project.metadata.name, pipeline_id=run_id
        )

        if context:
            project.notifiers.push_pipeline_start_message(
                project.metadata.name,
                context.uid,
            )
        else:
            project.push_pipeline_notification_kfp_runner(
                run_id,
                mlrun_pipelines.common.models.RunStatuses.running,
                f"Workflow {run_id} started in project {project.metadata.name}",
                notifications,
            )
        pipeline_context.clear()
        return _PipelineRunStatus(run_id, cls, project=project, workflow=workflow_spec)

    @classmethod
    def retry(
        cls,
        run: "_PipelineRunStatus",
        project: typing.Optional["mlrun.projects.MlrunProject"] = None,
    ) -> str:
        project_name = project.metadata.name if project else ""
        logger.info(
            "Retrying pipeline",
            run_id=run.run_id,
            project=project_name,
        )
        run_id = retry_pipeline(
            run.run_id,
            project=project_name,
        )
        return run_id

    @classmethod
    def terminate(
        cls,
        run: "_PipelineRunStatus",
        project: typing.Optional["mlrun.projects.MlrunProject"] = None,
    ) -> str:
        project_name = project.metadata.name if project else ""
        logger.info(
            "Terminating pipeline",
            run_id=run.run_id,
            project=project_name,
        )
        run_id = terminate_pipeline(
            run.run_id,
            project=project_name,
        )
        return run_id

    @staticmethod
    def wait_for_completion(
        run: "_PipelineRunStatus",
        project: typing.Optional["mlrun.projects.MlrunProject"] = None,
        timeout: typing.Optional[int] = None,
        expected_statuses: typing.Optional[list[str]] = None,
    ):
        project_name = project.metadata.name if project else ""
        logger.info(
            "Waiting for pipeline run completion",
            run_id=run.run_id,
            project=project_name,
        )
        timeout = timeout or 60 * 60
        run_info = wait_for_pipeline_completion(
            run.run_id,
            timeout=timeout,
            expected_statuses=expected_statuses,
            project=project_name,
        )
        status = ""
        if run_info:
            status = run_info["run"].get("status")
        return status

    @staticmethod
    def get_state(run_id, project=None):
        project_name = project.metadata.name if project else ""
        resp = mlrun.run.get_pipeline(run_id, project=project_name)
        if resp:
            return resp["run"].get("status", "")
        return ""


class _LocalRunner(_PipelineRunner):
    """local pipelines runner"""

    engine = "local"

    @classmethod
    def run(
        cls,
        project,
        workflow_spec: WorkflowSpec,
        name=None,
        workflow_handler=None,
        secrets=None,
        artifact_path=None,
        namespace=None,
        source=None,
        notifications: typing.Optional[list[mlrun.model.Notification]] = None,
        context: typing.Optional[mlrun.execution.MLClientCtx] = None,
    ) -> _PipelineRunStatus:
        pipeline_context.set(project, workflow_spec)
        workflow_handler = _PipelineRunner._get_handler(
            workflow_handler, workflow_spec, project, secrets
        )

        # fallback to old notification behavior
        for notification in notifications or []:
            project.notifiers.add_notification(notification.kind, notification.params)

        workflow_id = uuid.uuid4().hex
        pipeline_context.workflow_id = workflow_id
        # When using KFP, it would do this replacement. When running locally, we need to take care of it.
        if artifact_path:
            artifact_path = artifact_path.replace("{{workflow.uid}}", workflow_id)
        original_source = None
        if source:
            original_source = project.spec.source
            project.set_source(source=source)
        pipeline_context.workflow_artifact_path = artifact_path

        # TODO: we should create endpoint for sending custom notification from BE
        project.notifiers.push_pipeline_start_message_from_client(
            project.metadata.name, pipeline_id=workflow_id
        )
        err = None
        try:
            workflow_handler(**workflow_spec.args)
            state = mlrun_pipelines.common.models.RunStatuses.succeeded
        except Exception as exc:
            err = exc
            logger.exception("Workflow run failed")
            project.notifiers.push(
                f":x: Workflow {workflow_id} run failed!, error: {err_to_str(exc)}",
                mlrun.common.schemas.NotificationSeverity.ERROR,
            )
            state = mlrun_pipelines.common.models.RunStatuses.failed
        mlrun.run.wait_for_runs_completion(pipeline_context.runs_map.values())
        project.notifiers.push_pipeline_run_results(
            pipeline_context.runs_map.values(), state=state
        )
        pipeline_context.clear()

        # Setting the source back to the original in the project object
        if original_source:
            project.set_source(source=original_source)
        return _PipelineRunStatus(
            workflow_id,
            cls,
            project=project,
            workflow=workflow_spec,
            state=state,
            exc=err,
        )

    @staticmethod
    def get_state(run_id, project=None):
        return ""

    @staticmethod
    def wait_for_completion(run, project=None, timeout=None, expected_statuses=None):
        # TODO: local runner blocks for the duration of the pipeline.
        # Therefore usually there will be nothing to wait for.
        # However, users may run functions with watch=False and then it can be useful to wait for the runs here.
        pass


class _RemoteRunner(_PipelineRunner):
    """remote pipelines runner"""

    engine = "remote"

    @classmethod
    def run(
        cls,
        project: "mlrun.projects.MlrunProject",
        workflow_spec: WorkflowSpec,
        name: typing.Optional[str] = None,
        workflow_handler: typing.Optional[typing.Union[str, typing.Callable]] = None,
        secrets: mlrun.secrets.SecretsStore = None,
        artifact_path: typing.Optional[str] = None,
        namespace: typing.Optional[str] = None,
        source: typing.Optional[str] = None,
        notifications: typing.Optional[list[mlrun.model.Notification]] = None,
        context: typing.Optional[mlrun.execution.MLClientCtx] = None,
    ) -> typing.Optional[_PipelineRunStatus]:
        workflow_name = normalize_workflow_name(name=name, project_name=project.name)
        workflow_id = None

        # The returned engine for this runner is the engine of the workflow.
        # In this way wait_for_completion/get_run_status would be executed by the correct pipeline runner.
        inner_engine = get_workflow_engine(workflow_spec.engine)
        run_db = mlrun.get_run_db()
        err = None
        try:
            logger.info(
                "Submitting remote workflow",
                workflow_engine=workflow_spec.engine,
                schedule=workflow_spec.schedule,
                project_name=project.name,
            )

            # set it relative to project path
            # as the runner pod will mount and use `load_and_run` which will use the project context
            # to load the workflow file to.
            # e.g.
            # /path/to/project/workflow.py -> ./workflow.py
            # /path/to/project/subdir/workflow.py -> ./workflow.py
            if workflow_spec.path:
                prefix = project.spec.get_code_path()
                if workflow_spec.path.startswith(prefix):
                    workflow_spec.path = workflow_spec.path.removeprefix(prefix)
                    relative_prefix = "."
                    if not workflow_spec.path.startswith("/"):
                        relative_prefix += "/"
                    workflow_spec.path = f"{relative_prefix}{workflow_spec.path}"

            workflow_response = run_db.submit_workflow(
                project=project.name,
                name=workflow_name,
                workflow_spec=workflow_spec,
                artifact_path=artifact_path,
                source=source,
                run_name=config.workflows.default_workflow_runner_name.format(
                    workflow_name
                ),
                namespace=namespace,
                notifications=notifications,
            )
            if workflow_spec.schedule:
                logger.info(
                    "Workflow scheduled successfully",
                    workflow_response=workflow_response,
                )
                return

            get_workflow_id_timeout = max(
                int(mlrun.mlconf.workflows.timeouts.remote),
                int(getattr(mlrun.mlconf.workflows.timeouts, inner_engine.engine)),
            )

            logger.debug(
                "Workflow submitted, waiting for pipeline run to start",
                workflow_name=workflow_response.name,
                get_workflow_id_timeout=get_workflow_id_timeout,
            )

            def _get_workflow_id_or_bail():
                try:
                    return run_db.get_workflow_id(
                        project=project.name,
                        name=workflow_response.name,
                        run_id=workflow_response.run_id,
                        engine=workflow_spec.engine,
                    )
                except mlrun.errors.MLRunHTTPStatusError as get_wf_exc:
                    # fail fast on specific errors
                    if get_wf_exc.error_status_code in [
                        http.HTTPStatus.PRECONDITION_FAILED
                    ]:
                        raise mlrun.errors.MLRunFatalFailureError(
                            original_exception=get_wf_exc
                        )

                    # raise for a retry (on other errors)
                    raise

            # Getting workflow id from run:
            response = retry_until_successful(
                1,
                get_workflow_id_timeout,
                logger,
                False,
                _get_workflow_id_or_bail,
            )
            workflow_id = response.workflow_id
            # After fetching the workflow_id the workflow executed successfully

        except Exception as exc:
            err = exc
            logger.exception("Workflow run failed")
            project.notifiers.push(
                f":x: Workflow {workflow_name} run failed!, error: {err_to_str(exc)}",
                mlrun.common.schemas.NotificationSeverity.ERROR,
            )
            state = mlrun_pipelines.common.models.RunStatuses.failed
        else:
            state = mlrun_pipelines.common.models.RunStatuses.running
            pipeline_context.clear()
        return _PipelineRunStatus(
            run_id=workflow_id,
            engine=inner_engine,
            project=project,
            workflow=workflow_spec,
            state=state,
            exc=err,
        )

    @staticmethod
    def get_run_status(
        project,
        run: _PipelineRunStatus,
        timeout=None,
        expected_statuses=None,
        notifiers: mlrun.utils.notifications.CustomNotificationPusher = None,
        inner_engine: typing.Optional[type[_PipelineRunner]] = None,
    ):
        inner_engine = inner_engine or _KFPRunner
        if inner_engine.engine == _KFPRunner.engine:
            # ignore notifiers for remote notifications, as they are handled by the remote pipeline notifications,
            # so overriding with CustomNotificationPusher with empty list of notifiers or only local notifiers
            local_project_notifiers = list(
                set(mlrun.utils.notifications.NotificationTypes.local()).intersection(
                    set(project.notifiers.notifications.keys())
                )
            )
            notifiers = mlrun.utils.notifications.CustomNotificationPusher(
                local_project_notifiers
            )
            return _KFPRunner.get_run_status(
                project,
                run,
                timeout,
                expected_statuses,
                notifiers=notifiers,
            )

        elif inner_engine.engine == _LocalRunner.engine:
            mldb = mlrun.db.get_run_db(secrets=project._secrets)
            pipeline_runner_run = mldb.read_run(run.run_id, project=project.name)

            pipeline_runner_run = mlrun.run.RunObject.from_dict(pipeline_runner_run)

            # here we are waiting for the pipeline run to complete and refreshing after that the pipeline run from the
            # db
            # TODO: do it with timeout
            pipeline_runner_run.logs(db=mldb)
            pipeline_runner_run.refresh()
            run._state = mlrun.common.runtimes.constants.RunStates.run_state_to_pipeline_run_status(
                pipeline_runner_run.status.state
            )
            run._exc = pipeline_runner_run.status.error
            return _LocalRunner.get_run_status(
                project,
                run,
                timeout,
                expected_statuses,
                notifiers=notifiers,
            )

        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unsupported inner runner engine: {inner_engine.engine}"
            )


def create_pipeline(project, pipeline, functions, secrets=None, handler=None):
    spec = imputil.spec_from_file_location("workflow", pipeline)
    if spec is None:
        raise ImportError(f"Cannot import workflow {pipeline}")
    mod = imputil.module_from_spec(spec)
    spec.loader.exec_module(mod)

    setattr(mod, "funcs", functions)  # should be replaced with "functions" in future
    setattr(mod, "functions", functions)
    setattr(mod, "this_project", project)

    # verify all functions are in this project (init_functions may add new functions)
    for f in functions.values():
        f.metadata.project = project.metadata.name

    if not handler and hasattr(mod, "kfpipeline"):
        handler = "kfpipeline"
    if not handler and hasattr(mod, "pipeline"):
        handler = "pipeline"
    if not handler or not hasattr(mod, handler):
        raise ValueError(
            f"'workflow_handler' is not defined. "
            f"Either provide it as set_workflow argument, or include a function named"
            f" '{handler or 'pipeline'}' in your workflow .py file."
        )

    return getattr(mod, handler)


def github_webhook(request):
    signature = request.headers.get("X-Hub-Signature")
    data = request.data
    print("sig:", signature)
    print("headers:", request.headers)
    print("data:", data)
    print("json:", request.get_json())

    if request.headers.get("X-GitHub-Event") == "ping":
        return {"msg": "Ok"}

    return {"msg": "pushed"}


def rerun_workflow(
    context: mlrun.execution.MLClientCtx,
    run_uid: str,
    project_name: str,
    original_runner_uid: str,
    original_workflow_name: str,
):
    """
    Re-run a workflow by retrying a previously failed KFP pipeline.

    :param context:      MLRun context.
    :param run_uid:      The run UID of the original workflow to retry.
    :param project_name: The project name.
    :param original_runner_uid: The original workflow runner UID.
    :param original_workflow_name: The original workflow name.
    """
    db = mlrun.get_run_db()
    new_pipeline_id = None

    try:
        # Invoke the KFP retry endpoint (direct-submit mode)
        new_pipeline_id = db.retry_pipeline(
            run_id=run_uid,
            project=project_name,
            submit_mode=mlrun_constants.WorkflowSubmitMode.direct,
        )
        logger.info(
            "KFP retry submitted",
            new_pipeline_id=new_pipeline_id,
            rerun_of_workflow=run_uid,
        )

        # Enqueue "running" notifications server-side for this RerunRunner run
        db.push_run_notifications(context.uid, project_name)

        context.set_label(
            mlrun_constants.MLRunInternalLabels.workflow_id, new_pipeline_id
        )
        context.update_run()

        context.log_result("workflow_id", new_pipeline_id)

        pipeline = wait_for_pipeline_completion(
            new_pipeline_id,
            project=project_name,
        )

        final_state = pipeline["run"]["status"]
        context.log_result("workflow_state", final_state, commit=True)

    except mlrun.errors.MLRunHTTPError as http_exc:
        logger.error(
            "Failed calling KFP retry API",
            run_id=run_uid,
            error=err_to_str(http_exc),
        )
        raise

    except Exception as exc:
        logger.error(
            "Error during rerun_workflow execution",
            error=err_to_str(exc),
            rerun_pipeline_id=new_pipeline_id,
        )
        raise

    finally:
        # Once the rerun has finished, clear the “retrying” label on the original runner
        # so that subsequent retry requests can acquire the lock again.
        db.set_run_retrying_status(
            project=project_name,
            name=original_workflow_name,
            run_id=original_runner_uid,
            retrying=False,
        )

    if final_state != mlrun_pipelines.common.models.RunStatuses.succeeded:
        raise mlrun.errors.MLRunRuntimeError(
            f"Pipeline retry of {run_uid} finished in state={final_state}"
        )


def load_and_run(context, *args, **kwargs):
    """
    This function serves as an alias to `load_and_run_workflow`,
    allowing to continue using `load_and_run` without modifying existing workflows or exported runs.
    This approach ensures backward compatibility,
    while directing all new calls to the updated `load_and_run_workflow` function.
    """
    kwargs.pop("load_only", None)
    kwargs.pop("save", None)
    load_and_run_workflow(context, *args, **kwargs)


def load_and_run_workflow(
    context: mlrun.execution.MLClientCtx,
    url: typing.Optional[str] = None,
    project_name: str = "",
    init_git: typing.Optional[bool] = None,
    subpath: typing.Optional[str] = None,
    clone: bool = False,
    workflow_name: typing.Optional[str] = None,
    workflow_path: typing.Optional[str] = None,
    workflow_arguments: typing.Optional[dict[str, typing.Any]] = None,
    artifact_path: typing.Optional[str] = None,
    workflow_handler: typing.Optional[typing.Union[str, typing.Callable]] = None,
    namespace: typing.Optional[str] = None,
    sync: bool = False,
    dirty: bool = False,
    engine: typing.Optional[str] = None,
    local: typing.Optional[bool] = None,
    schedule: typing.Union[str, mlrun.common.schemas.ScheduleCronTrigger] = None,
    cleanup_ttl: typing.Optional[int] = None,
    wait_for_completion: bool = False,
    project_context: typing.Optional[str] = None,
):
    """
    Auxiliary function that the RemoteRunner run once or run every schedule.
    This function loads a project from a given remote source and then runs the workflow.
    :param context:             mlrun context.
    :param url:                 remote url that represents the project's source.
                                See 'mlrun.load_project()' for details
    :param project_name:        project name
    :param init_git:            if True, will git init the context dir
    :param subpath:             project subpath (within the archive)
    :param clone:               if True, always clone (delete any existing content)
    :param workflow_name:       name of the workflow
    :param workflow_path:       url to a workflow file, if not a project workflow
    :param workflow_arguments:  kubeflow pipelines arguments (parameters)
    :param artifact_path:       target path/url for workflow artifacts, the string
                                '{{workflow.uid}}' will be replaced by workflow id
    :param workflow_handler:    workflow function handler (for running workflow function directly)
    :param namespace:           kubernetes namespace if other than default
    :param sync:                force functions sync before run
    :param dirty:               allow running the workflow when the git repo is dirty
    :param engine:              workflow engine running the workflow.
                                supported values are 'kfp' (default) or 'local'
    :param local:               run local pipeline with local functions (set local=True in function.run())
    :param schedule:            ScheduleCronTrigger class instance or a standard crontab expression string
    :param cleanup_ttl:         pipeline cleanup ttl in secs (time to wait after workflow completion, at which point the
                                workflow and all its resources are deleted)
    :param wait_for_completion: wait for workflow completion before returning
    :param project_context:     project context path (used for loading the project)
    """
    project_context = project_context or f"./{project_name}"

    # Load the project to fetch files which the runner needs, such as remote source files
    pull_remote_project_files(
        context=context,
        project_context=project_context,
        url=url,
        project_name=project_name,
        init_git=init_git,
        subpath=subpath,
        clone=clone,
        schedule=schedule,
        workflow_name=workflow_name,
    )

    # Retrieve the project object:
    # - If the project exists in the MLRun database, it will be loaded from there.
    # - If it doesn't exist in the database, it will be created from the previously loaded local directory.
    project = mlrun.get_or_create_project(
        context=project_context or f"./{project_name}",
        name=project_name,
        allow_cross_project=True,
    )

    # extract "start" notification if exists
    start_notifications = [
        notification
        for notification in context.get_notifications(unmask_secret_params=True)
        if mlrun.common.runtimes.constants.RunStates.running in notification.when
    ]

    # Prevent redundant notifications for run completion by ensuring that notifications are only triggered when the run
    # reaches the "running" state, as the server already handles the completion notifications.
    for notification in start_notifications:
        notification.when = [mlrun.common.runtimes.constants.RunStates.running]

    workflow_log_message = workflow_name or workflow_path
    context.logger.info(
        "Running workflow from remote", workflow_log_message=workflow_log_message
    )
    run = project.run(
        name=workflow_name,
        workflow_path=workflow_path,
        arguments=workflow_arguments,
        artifact_path=artifact_path,
        workflow_handler=workflow_handler,
        namespace=namespace,
        sync=sync,
        watch=False,  # Required for fetching the workflow_id
        dirty=dirty,
        cleanup_ttl=cleanup_ttl,
        engine=engine,
        local=local,
        notifications=start_notifications,
        context=context,
    )
    # Patch the current run object (the workflow-runner) with the workflow-id label
    context.logger.info(
        "Associating workflow-runner with workflow ID", run_id=run.run_id
    )
    context.set_label(mlrun_constants.MLRunInternalLabels.workflow_id, run.run_id)
    context.update_run()

    context.log_result(key="workflow_id", value=run.run_id)
    context.log_result(key="engine", value=run._engine.engine, commit=True)

    if run.state == mlrun_pipelines.common.models.RunStatuses.failed:
        raise RuntimeError(f"Workflow {workflow_log_message} failed") from run.exc

    if wait_for_completion:
        handle_workflow_completion(
            run=run,
            project=project,
            context=context,
            workflow_log_message=workflow_log_message,
        )


def pull_remote_project_files(
    context: mlrun.execution.MLClientCtx,
    project_context: str,
    url: str,
    project_name: str,
    init_git: typing.Optional[bool],
    subpath: typing.Optional[str],
    clone: bool,
    schedule: typing.Optional[
        typing.Union[str, mlrun.common.schemas.ScheduleCronTrigger]
    ],
    workflow_name: typing.Optional[str],
) -> None:
    """
    Load the project to clone remote files if they exist.
    If an exception occurs during project loading, send a notification if the workflow is scheduled.

    :param context:        MLRun execution context.
    :param project_context: Path to the project context.
    :param url:            URL of the project repository.
    :param project_name:   Name of the project.
    :param init_git:       Initialize a git repository.
    :param subpath:        Project subpath within the repository.
    :param clone:          Whether to clone the repository.
    :param schedule:       Schedule for running the workflow.
    :param workflow_name:  Name of the workflow to run.
    """
    try:
        # Load the project to clone remote files if they exist.
        # Using save=False to avoid overriding changes from the database if it already exists.
        mlrun.load_project(
            context=project_context,
            url=url,
            name=project_name,
            init_git=init_git,
            subpath=subpath,
            clone=clone,
            save=False,
            allow_cross_project=True,
        )
    except Exception as error:
        notify_scheduled_workflow_failure(
            schedule=schedule,
            project_name=project_name,
            workflow_name=workflow_name,
            error=error,
            context_uid=context.uid,
        )
        raise error


def notify_scheduled_workflow_failure(
    schedule,
    project_name: str,
    workflow_name: str,
    error: Exception,
    context_uid: str,
) -> None:
    if schedule:
        notification_pusher = mlrun.utils.notifications.CustomNotificationPusher(
            ["slack"]
        )
        url = get_workflow_url(project_name, context_uid)
        link = f"<{url}|*view workflow job details*>"
        message = (
            f":x: Failed to run scheduled workflow {workflow_name} "
            f"in Project {project_name}!\n"
            f"Error: ```{err_to_str(error)}```\n{link}"
        )
        # Sending Slack Notification without losing the original error:
        try:
            notification_pusher.push(
                message=message,
                severity=mlrun.common.schemas.NotificationSeverity.ERROR,
            )

        except Exception as exc:
            logger.error("Failed to send slack notification", exc=err_to_str(exc))


def handle_workflow_completion(
    run: _PipelineRunStatus,
    project,
    context: mlrun.execution.MLClientCtx,
    workflow_log_message: str,
) -> None:
    """
    Handle workflow completion by waiting for it to finish and logging the final state.

    :param run:                 Run object containing workflow execution details.
    :param project:             MLRun project object.
    :param context:             MLRun execution context.
    :param workflow_log_message: Message used for logging.
    """
    try:
        run.wait_for_completion()
    except Exception as exc:
        mlrun.utils.logger.error(
            "Failed waiting for workflow completion",
            workflow=workflow_log_message,
            exc=err_to_str(exc),
        )

    pipeline_state, _, _ = project.get_run_status(run)
    context.log_result(key="workflow_state", value=pipeline_state, commit=True)
    if pipeline_state != mlrun_pipelines.common.models.RunStatuses.succeeded:
        raise RuntimeError(
            f"Workflow {workflow_log_message} failed, state={pipeline_state}"
        )


def import_remote_project(
    context: mlrun.execution.MLClientCtx,
    url: typing.Optional[str] = None,
    project_name: str = "",
    init_git: typing.Optional[bool] = None,
    subpath: typing.Optional[str] = None,
    clone: bool = False,
    save: bool = True,
    project_context: typing.Optional[str] = None,
):
    """
    This function loads a project from a given remote source.

    :param context:             mlrun context.
    :param url:                 remote url that represents the project's source.
                                See 'mlrun.load_project()' for details
    :param project_name:        project name
    :param init_git:            if True, will git init the context dir
    :param subpath:             project subpath (within the archive)
    :param clone:               if True, always clone (delete any existing content)
    :param save:                whether to save the created project and artifact in the DB
    :param project_context:     project context path (used for loading the project)
    """
    project = mlrun.load_project(
        context=project_context or f"./{project_name}",
        url=url,
        name=project_name,
        init_git=init_git,
        subpath=subpath,
        clone=clone,
        save=save,
        sync_functions=True,
    )

    context.logger.info("Loaded project successfully", project_name=project.name)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from abc import ABC
from tempfile import NamedTemporaryFile
from typing import Optional

import semver

import mlrun.model_monitoring.applications.base as mm_base
import mlrun.model_monitoring.applications.context as mm_context
from mlrun.errors import MLRunIncompatibleVersionError, MLRunValueError

SUPPORTED_EVIDENTLY_VERSION = semver.Version.parse("0.7.5")


def _check_evidently_version(*, cur: semver.Version, ref: semver.Version) -> None:
    if ref.is_compatible(cur) or (
        cur.major == ref.major == 0 and cur.minor == ref.minor and cur.patch > ref.patch
    ):
        return
    if cur.major == ref.major == 0 and cur.minor > ref.minor:
        warnings.warn(
            f"Evidently version {cur} is not compatible with the tested "
            f"version {ref}, use at your own risk."
        )
    else:
        raise MLRunIncompatibleVersionError(
            f"Evidently version {cur} is not supported, please change to "
            f"{ref} (or another compatible version)."
        )


_HAS_EVIDENTLY = False
try:
    import evidently  # noqa: F401

    _check_evidently_version(
        cur=semver.Version.parse(evidently.__version__),
        ref=SUPPORTED_EVIDENTLY_VERSION,
    )
    _HAS_EVIDENTLY = True
except ModuleNotFoundError:
    pass


if _HAS_EVIDENTLY:
    from evidently.core.report import Snapshot
    from evidently.ui.workspace import (
        STR_UUID,
        CloudWorkspace,
        Project,
        Workspace,
        WorkspaceBase,
    )


class EvidentlyModelMonitoringApplicationBase(
    mm_base.ModelMonitoringApplicationBase, ABC
):
    def __init__(
        self,
        evidently_project_id: "STR_UUID",
        evidently_workspace_path: Optional[str] = None,
        cloud_workspace: bool = False,
    ) -> None:
        """
        A class for integrating Evidently for MLRun model monitoring within a monitoring application.

        .. note::

            The ``evidently`` package is not installed by default in the mlrun/mlrun image.
            It must be installed separately to use this class.

        :param evidently_project_id:        (str) The ID of the Evidently project.
        :param evidently_workspace_path:    (str) The path to the Evidently workspace.
        :param cloud_workspace:             (bool) Whether the workspace is an Evidently Cloud workspace.
        """
        if not _HAS_EVIDENTLY:
            raise ModuleNotFoundError("Evidently is not installed - the app cannot run")
        self.evidently_workspace_path = evidently_workspace_path
        if cloud_workspace:
            self.get_workspace = self.get_cloud_workspace
        self.evidently_workspace = self.get_workspace()
        self.evidently_project_id = evidently_project_id
        self.evidently_project = self.load_project()

    def load_project(self) -> "Project":
        """Load the Evidently project."""
        return self.evidently_workspace.get_project(self.evidently_project_id)

    def get_workspace(self) -> "WorkspaceBase":
        """Get the Evidently workspace. Override this method for customize access to the workspace."""
        if self.evidently_workspace_path:
            return Workspace.create(self.evidently_workspace_path)
        else:
            raise MLRunValueError(
                "A local workspace could not be created as `evidently_workspace_path` is not set.\n"
                "If you intend to use a cloud workspace, please use `cloud_workspace=True` and set the "
                "`EVIDENTLY_API_KEY` environment variable. In other cases, override this method."
            )

    def get_cloud_workspace(self) -> "CloudWorkspace":
        """Load the Evidently cloud workspace according to the `EVIDENTLY_API_KEY` environment variable."""
        return CloudWorkspace()

    @staticmethod
    def log_evidently_object(
        monitoring_context: mm_context.MonitoringApplicationContext,
        evidently_object: "Snapshot",
        artifact_name: str,
        unique_per_endpoint: bool = True,
    ) -> None:
        """
        Logs an Evidently report or suite as an artifact.

        .. caution::

            Logging Evidently objects in every model monitoring window may cause scale issues.
            This method should be called on special occasions only.

        :param monitoring_context:  (MonitoringApplicationContext) The monitoring context to process.
        :param evidently_object:    (Snapshot) The Evidently run to log, e.g. a report run.
        :param artifact_name:       (str) The name for the logged artifact.
        :param unique_per_endpoint: by default ``True``, we will log different artifact for each model endpoint,
                                    set to ``False`` without changing item key will cause artifact override.
        """
        with NamedTemporaryFile(suffix=".html") as file:
            evidently_object.save_html(filename=file.name)
            monitoring_context.log_artifact(
                artifact_name,
                local_path=file.name,
                unique_per_endpoint=unique_per_endpoint,
            )



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import mlrun.errors


class ObjectFormat:
    """
    MLRun object formatter. Any class that inherits from this class should implement the `format_method` method
    to specify the formatting method for each format.
    A `filter_obj_method` utility method is provided to filter the object based on a list of keys.
    """

    full = "full"

    @staticmethod
    def format_method(format_: str) -> typing.Optional[typing.Callable]:
        """
        Get the formatting method for the provided format.
        A `None` value signifies a pass-through formatting method (no formatting).
        :param format_: The format as a string representation.
        :return: The formatting method.
        """
        return {
            ObjectFormat.full: None,
        }[format_]

    @classmethod
    def format_obj(
        cls,
        obj: typing.Any,
        format_: str,
        exclude_formats: typing.Optional[list[str]] = None,
    ) -> typing.Any:
        """
        Format the provided object based on the provided format.
        :param obj: The object to format.
        :param format_: The format as a string representation.
        :param exclude_formats: A list of formats to exclude from the formatting process. If the provided format is in
                                this list, an invalid format exception will be raised.
        """
        exclude_formats = exclude_formats or []
        format_ = format_ or cls.full
        invalid_format_exc = mlrun.errors.MLRunBadRequestError(
            f"Provided format is not supported. format={format_}"
        )

        if format_ in exclude_formats:
            raise invalid_format_exc

        try:
            format_method = cls.format_method(format_)
        except KeyError:
            raise invalid_format_exc

        if not format_method:
            return obj

        return format_method(obj)

    @staticmethod
    def filter_obj_method(_filter: list[str]) -> typing.Callable:
        """
        Returns a method that filters the object based on the provided list of keys.
        The keys should be in a dot-separated format, denoting the path within the dictionary to the desired key.
        The object maintains its structure, with the filtered keys and their values, while all other keys are removed.
        :param _filter: The list of keys to filter by.
                        Example:
                        [
                            "kind",
                            "metadata.name",
                            "spec.something.else",
                        ]

        :return: The filtering method.
        """

        def _filter_method(obj: dict) -> dict:
            formatted_obj = {}
            for key in _filter:
                key_list = key.split(".")
                obj_recursive_iterator = obj
                formatted_obj_recursive_iterator = formatted_obj
                for idx, key in enumerate(key_list):
                    if key not in obj_recursive_iterator:
                        break
                    value = (
                        {} if idx < len(key_list) - 1 else obj_recursive_iterator[key]
                    )
                    formatted_obj_recursive_iterator.setdefault(key, value)

                    obj_recursive_iterator = obj_recursive_iterator[key]
                    formatted_obj_recursive_iterator = formatted_obj_recursive_iterator[
                        key
                    ]

            return formatted_obj

        return _filter_method



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def my_func(context, p1: int = 1, p2="a-string"):
    """this is a two param function

    :param p1  first param
    :param p2  2nd param
    """
    # access input metadata, values, files, and secrets (passwords)
    print(f"Run: {context.name} (uid={context.uid})")
    print(f"Params: p1={p1}, p2={p2}")
    context.logger.info("running function")

    # RUN some useful code e.g. ML training, data prep, etc.

    # log scalar result values (job result metrics)
    context.log_result("accuracy", p1 * 2)
    context.log_result("loss", p1 * 3)
    context.set_label("framework", "sklearn")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mlrun

from ..._common import LoggingMode
from ..utils import DLTypes


class Logger:
    """
    Logger for tracking hyperparamters and metrics results during training / evaluation of some framework.
    """

    def __init__(self, context: mlrun.MLClientCtx = None):
        """
        Initialize a generic logger for collecting training / validation runs data.

        :param context: MLRun context to log its parameters as static hyperparameters if needed.
        """
        # Save the context:
        self._context = context

        # Set up the logger's mode (default:  Training):
        self._mode = LoggingMode.TRAINING

        # Set up the results dictionaries - a dictionary of metrics for all the iteration results by their epochs:
        # [Metric: str] -> [Epoch: int] -> [Iteration: int] -> [value: float]
        self._training_results = {}  # type: Dict[str, List[List[float]]]
        self._validation_results = {}  # type: Dict[str, List[List[float]]]

        # Set up the metrics summary dictionaries - a dictionary of all metrics averages by epochs:
        # [Metric: str] -> [Epoch: int] -> [value: float]:
        self._training_summaries = {}  # type: Dict[str, List[float]]
        self._validation_summaries = {}  # type: Dict[str, List[float]]

        # Store the static hyperparameters given - a dictionary of parameters and their values to note:
        # [Parameter: str] -> [value: Union[str, bool, float, int]]
        self._static_hyperparameters = {}  # type: Dict[str, DLTypes.TrackableType]

        # Setup the dynamic hyperparameters dictionary - a dictionary of all tracked hyperparameters by epochs:
        # [Hyperparameter: str] -> [Epoch: int] -> [value: Union[str, bool, float, int]]
        self._dynamic_hyperparameters = {}  # type: Dict[str, List[DLTypes.TrackableType]]

        # Setup the iterations counter:
        self._epochs = 0
        self._training_iterations = 0
        self._validation_iterations = 0

    @property
    def context(self) -> mlrun.MLClientCtx:
        """
        Get the loggers context. If its not set, None will be returned.

        :return: The logger's context.
        """
        return self._context

    @property
    def mode(self) -> LoggingMode:
        """
        Get the logger's mode.

        :return: The logger's mode. One of Logger.LoggerMode.
        """
        return self._mode

    @property
    def training_results(self) -> dict[str, list[list[float]]]:
        """
        Get the training results logged. The results will be stored in a dictionary where each key is the metric name
        and the value is a list of lists of values. The first list is by epoch and the second list is by iteration
        (batch).

        :return: The training results.
        """
        return self._training_results

    @property
    def validation_results(self) -> dict[str, list[list[float]]]:
        """
        Get the validation results logged. The results will be stored in a dictionary where each key is the metric name
        and the value is a list of lists of values. The first list is by epoch and the second list is by iteration
        (batch).

        :return: The validation results.
        """
        return self._validation_results

    @property
    def training_summaries(self) -> dict[str, list[float]]:
        """
        Get the training summaries of the metrics results. The summaries will be stored in a dictionary where each key
        is the metric names and the value is a list of all the summary values per epoch.

        :return: The training summaries.
        """
        return self._training_summaries

    @property
    def validation_summaries(self) -> dict[str, list[float]]:
        """
        Get the validation summaries of the metrics results. The summaries will be stored in a dictionary where each key
        is the metric names and the value is a list of all the summary values per epoch.

        :return: The validation summaries.
        """
        return self._validation_summaries

    @property
    def static_hyperparameters(self) -> dict[str, DLTypes.TrackableType]:
        """
        Get the static hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is his logged value.

        :return: The static hyperparameters.
        """
        return self._static_hyperparameters

    @property
    def dynamic_hyperparameters(self) -> dict[str, list[DLTypes.TrackableType]]:
        """
        Get the dynamic hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is a list of his logged values per epoch.

        :return: The dynamic hyperparameters.
        """
        return self._dynamic_hyperparameters

    @property
    def epochs(self) -> int:
        """
        Get the overall epochs.

        :return: The overall epochs.
        """
        return self._epochs

    @property
    def training_iterations(self) -> int:
        """
        Get the overall training iterations.

        :return: The overall training iterations.
        """
        return self._training_iterations

    @property
    def validation_iterations(self) -> int:
        """
        Get the overall validation iterations.

        :return: The overall validation iterations.
        """
        return self._validation_iterations

    def set_mode(self, mode: LoggingMode):
        """
        Set the logger's mode.

        :param mode: The mode to set. One of Logger.LoggerMode.
        """
        self._mode = mode

    def log_epoch(self):
        """
        Log a new epoch, appending all the result with a new list for the new epoch.
        """
        # Count the new epoch:
        self._epochs += 1

        # Add a new epoch to each of the metrics in the results dictionary:
        for results_dictionary in [self._training_results, self._validation_results]:
            for metric in results_dictionary:
                results_dictionary[metric].append([])

    def log_training_iteration(self):
        """
        Log a new training iteration.
        """
        self._training_iterations += 1

    def log_validation_iteration(self):
        """
        Log a new validation iteration.
        """
        self._validation_iterations += 1

    def log_training_result(self, metric_name: str, result: float):
        """
        Log the given metric result in the training results dictionary at the current epoch.

        :param metric_name: The metric name as it was logged in 'log_metric'.
        :param result:      The metric result to log.
        """
        if metric_name not in self._training_results:
            self._training_results[metric_name] = [[]]
        self._training_results[metric_name][-1].append(result)

    def log_validation_result(self, metric_name: str, result: float):
        """
        Log the given metric result in the validation results dictionary at the current epoch.

        :param metric_name: The metric name as it was logged in 'log_metric'.
        :param result:      The metric result to log.
        """
        if metric_name not in self._validation_results:
            self._validation_results[metric_name] = [[]]
        self._validation_results[metric_name][-1].append(result)

    def log_training_summary(self, metric_name: str, result: float):
        """
        Log the given metric result in the training summaries results dictionary.

        :param metric_name: The metric name as it was logged in 'log_metric'.
        :param result:      The metric result to log.
        """
        if metric_name not in self._training_summaries:
            self._training_summaries[metric_name] = []
        self._training_summaries[metric_name].append(result)

    def log_validation_summary(self, metric_name: str, result: float):
        """
        Log the given metric result in the validation summaries results dictionary.

        :param metric_name: The metric name as it was logged in 'log_metric'.
        :param result:      The metric result to log.
        """
        if metric_name not in self._validation_summaries:
            self._validation_summaries[metric_name] = []
        self._validation_summaries[metric_name].append(result)

    def log_static_hyperparameter(
        self, parameter_name: str, value: DLTypes.TrackableType
    ):
        """
        Log the given parameter value in the static hyperparameters dictionary.

        :param parameter_name: The parameter name.
        :param value:          The parameter value to log.
        """
        self._static_hyperparameters[parameter_name] = value

    def log_dynamic_hyperparameter(
        self, parameter_name: str, value: DLTypes.TrackableType
    ):
        """
        Log the given parameter value in the dynamic hyperparameters dictionary at the current epoch (if its a new
        parameter it will be epoch 0). If the parameter appears in the static hyperparameters dictionary, it will be
        removed from there as it is now dynamic.

        :param parameter_name: The parameter name.
        :param value:          The parameter value to log.
        """
        # Check if its a new hyperparameter being tracked:
        if parameter_name not in self._dynamic_hyperparameters:
            # Look in the static hyperparameters:
            if parameter_name in self._static_hyperparameters:
                self._static_hyperparameters.pop(parameter_name)
            # Add it as a dynamic hyperparameter:
            self._dynamic_hyperparameters[parameter_name] = [value]
        else:
            self._dynamic_hyperparameters[parameter_name].append(value)

    # TODO: Move to MLRun logger
    def log_context_parameters(self):
        """
        Log the context given parameters as static hyperparameters. Should be called once as the context parameters do
        not change.
        """
        for parameter_name, parameter_value in self._context.parameters.items():
            # Check if the parameter is a trackable value:
            if isinstance(parameter_value, (str, bool, float, int)):
                self.log_static_hyperparameter(
                    parameter_name=parameter_name, value=parameter_value
                )
            else:
                # See if its string representation length is below the maximum value length:
                string_value = str(parameter_value)
                if (
                    len(string_value) < 30
                ):  # Temporary to no log to long variables into the UI.
                    # TODO: Make the user specify the parameters and take them all by default.
                    self.log_static_hyperparameter(
                        parameter_name=parameter_name, value=parameter_value
                    )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import re
from dataclasses import dataclass, field
from enum import Enum, IntEnum
from typing import Optional

import mlrun.common.constants
import mlrun.common.helpers
from mlrun.common.types import StrEnum


class MonitoringStrEnum(StrEnum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class ModelEndpointSchema(MonitoringStrEnum):
    # metadata
    UID = "uid"
    PROJECT = "project"
    ENDPOINT_TYPE = "endpoint_type"
    MODE = "mode"
    NAME = "name"
    CREATED = "created"
    UPDATED = "updated"
    LABELS = "labels"

    # spec
    FUNCTION_NAME = "function_name"
    FUNCTION_TAG = "function_tag"
    MODEL_NAME = "model_name"
    MODEL_TAGS = "model_tags"
    MODEL_PATH = "model_path"
    MODEL_CLASS = "model_class"
    FEATURE_NAMES = "feature_names"
    LABEL_NAMES = "label_names"
    FEATURE_STATS = "feature_stats"
    MONITORING_FEATURE_SET_URI = "monitoring_feature_set_uri"
    CHILDREN = "children"
    CHILDREN_UIDS = "children_uids"
    FUNCTION_URI = "function_uri"
    MODEL_URI = "model_uri"

    # status
    STATE = "state"
    MONITORING_MODE = "monitoring_mode"
    FIRST_REQUEST = "first_request"
    SAMPLING_PERCENTAGE = "sampling_percentage"

    # status - operative
    LAST_REQUEST = "last_request"
    RESULT_STATUS = "result_status"
    AVG_LATENCY = "avg_latency"
    ERROR_COUNT = "error_count"
    CURRENT_STATS = "current_stats"
    DRIFT_MEASURES = "drift_measures"


class ModelEndpointCreationStrategy(MonitoringStrEnum):
    INPLACE = "inplace"
    ARCHIVE = "archive"
    OVERWRITE = "overwrite"
    SKIP = "skip"


class EventFieldType:
    FUNCTION_URI = "function_uri"
    FUNCTION = "function"
    MODEL_URI = "model_uri"
    MODEL = "model"
    VERSION = "version"
    VERSIONED_MODEL = "versioned_model"
    MODEL_CLASS = "model_class"
    TIMESTAMP = "timestamp"
    # `endpoint_id` is deprecated as a field in the model endpoint schema since 1.3.1, replaced by `uid`.
    ENDPOINT_ID = "endpoint_id"
    ENDPOINT_NAME = "endpoint_name"
    UID = "uid"
    ENDPOINT_TYPE = "endpoint_type"
    REQUEST_ID = "request_id"
    RECORD_TYPE = "record_type"
    FEATURES = "features"
    FEATURE_NAMES = "feature_names"
    NAMED_FEATURES = "named_features"
    LABELS = "labels"
    LATENCY = "latency"
    LABEL_NAMES = "label_names"
    PREDICTION = "prediction"
    PREDICTIONS = "predictions"
    NAMED_PREDICTIONS = "named_predictions"
    ERROR_COUNT = "error_count"
    MODEL_ERROR = "model_error"
    ERROR_TYPE = "error_type"
    INFER_ERROR = "infer_error"
    ENTITIES = "entities"
    FIRST_REQUEST = "first_request"
    LAST_REQUEST = "last_request"
    LAST_REQUEST_TIMESTAMP = "last_request_timestamp"
    METRIC = "metric"
    METRICS = "metrics"
    BATCH_INTERVALS_DICT = "batch_intervals_dict"
    DEFAULT_BATCH_INTERVALS = "default_batch_intervals"
    MINUTES = "minutes"
    HOURS = "hours"
    DAYS = "days"
    MODEL_ENDPOINTS = "model_endpoints"
    STATE = "state"
    PROJECT = "project"
    STREAM_PATH = "stream_path"
    ACTIVE = "active"
    MONITORING_MODE = "monitoring_mode"
    FEATURE_STATS = "feature_stats"
    CURRENT_STATS = "current_stats"
    CHILDREN = "children"
    CHILDREN_UIDS = "children_uids"
    DRIFT_MEASURES = "drift_measures"
    DRIFT_STATUS = "drift_status"
    MONITOR_CONFIGURATION = "monitor_configuration"
    FEATURE_SET_URI = "monitoring_feature_set_uri"
    ALGORITHM = "algorithm"
    VALUE = "value"
    SAMPLE_PARQUET_PATH = "sample_parquet_path"
    TIME = "time"
    TABLE_COLUMN = "table_column"
    SAMPLING_PERCENTAGE = "sampling_percentage"
    SAMPLING_RATE = "sampling_rate"
    ESTIMATED_PREDICTION_COUNT = "estimated_prediction_count"
    EFFECTIVE_SAMPLE_COUNT = "effective_sample_count"


class StreamProcessingEvent:
    MODEL = "model"
    MODEL_CLASS = "model_class"
    MICROSEC = "microsec"
    WHEN = "when"
    ERROR = "error"
    ENDPOINT_ID = "endpoint_id"
    SAMPLING_PERCENTAGE = "sampling_percentage"
    EFFECTIVE_SAMPLE_COUNT = "effective_sample_count"
    LABELS = "labels"
    FUNCTION_URI = "function_uri"
    REQUEST = "request"
    RESPONSE = "resp"
    METRICS = "metrics"


class FeatureSetFeatures(MonitoringStrEnum):
    LATENCY = EventFieldType.LATENCY
    METRICS = EventFieldType.METRICS

    @classmethod
    def time_stamp(cls):
        return EventFieldType.TIMESTAMP

    @classmethod
    def entity(cls):
        return EventFieldType.ENDPOINT_ID


class ApplicationEvent:
    APPLICATION_NAME = "application_name"
    START_INFER_TIME = "start_infer_time"
    END_INFER_TIME = "end_infer_time"
    ENDPOINT_ID = "endpoint_id"
    ENDPOINT_NAME = "endpoint_name"
    ENDPOINT_UPDATED = "endpoint_updated"


class WriterEvent(MonitoringStrEnum):
    ENDPOINT_NAME = "endpoint_name"
    APPLICATION_NAME = "application_name"
    ENDPOINT_ID = "endpoint_id"
    START_INFER_TIME = "start_infer_time"
    END_INFER_TIME = "end_infer_time"
    EVENT_KIND = "event_kind"  # metric or result or stats
    DATA = "data"


class WriterEventKind(MonitoringStrEnum):
    METRIC = "metric"
    RESULT = "result"
    STATS = "stats"

    @classmethod
    def user_app_outputs(cls):
        return [cls.METRIC, cls.RESULT]


class ControllerEvent(MonitoringStrEnum):
    KIND = "kind"
    ENDPOINT_ID = "endpoint_id"
    ENDPOINT_NAME = "endpoint_name"
    PROJECT = "project"
    TIMESTAMP = "timestamp"
    FIRST_REQUEST = "first_request"
    FEATURE_SET_URI = "feature_set_uri"
    ENDPOINT_TYPE = "endpoint_type"

    # first_timestamp and last_timestamp are used to batch completed events
    FIRST_TIMESTAMP = "first_timestamp"
    LAST_TIMESTAMP = "last_timestamp"

    ENDPOINT_POLICY = "endpoint_policy"
    # Note: currently under endpoint policy we will have a dictionary including the keys: "application_names"
    # "base_period", and "updated_endpoint" stand for when the MEP was updated


class ControllerEventEndpointPolicy(MonitoringStrEnum):
    BASE_PERIOD = "base_period"
    MONITORING_APPLICATIONS = "monitoring_applications"
    ENDPOINT_UPDATED = "endpoint_updated"


class ControllerEventKind(MonitoringStrEnum):
    NOP_EVENT = "nop_event"
    REGULAR_EVENT = "regular_event"
    BATCH_COMPLETE = "batch_complete"


class MetricData(MonitoringStrEnum):
    METRIC_NAME = "metric_name"
    METRIC_VALUE = "metric_value"


class ResultData(MonitoringStrEnum):
    RESULT_NAME = "result_name"
    RESULT_VALUE = "result_value"
    RESULT_KIND = "result_kind"
    RESULT_STATUS = "result_status"
    RESULT_EXTRA_DATA = "result_extra_data"


class StatsData(MonitoringStrEnum):
    STATS_NAME = "stats_name"
    STATS = "stats"
    TIMESTAMP = "timestamp"


class StatsKind(MonitoringStrEnum):
    CURRENT_STATS = "current_stats"
    DRIFT_MEASURES = "drift_measures"


class EventLiveStats:
    LATENCY_AVG_5M = "latency_avg_5m"
    LATENCY_AVG_1H = "latency_avg_1h"
    PREDICTIONS_PER_SECOND = "predictions_per_second"
    PREDICTIONS_COUNT_5M = "predictions_count_5m"
    PREDICTIONS_COUNT_1H = "predictions_count_1h"


class EventKeyMetrics:
    BASE_METRICS = "base_metrics"
    CUSTOM_METRICS = "custom_metrics"
    ENDPOINT_FEATURES = "endpoint_features"
    GENERIC = "generic"
    REAL_TIME = "real_time"


class TSDBTarget(MonitoringStrEnum):
    V3IO_TSDB = "v3io-tsdb"
    TDEngine = "tdengine"


class ProjectSecretKeys:
    ACCESS_KEY = "MODEL_MONITORING_ACCESS_KEY"
    TSDB_PROFILE_NAME = "TSDB_PROFILE_NAME"
    STREAM_PROFILE_NAME = "STREAM_PROFILE_NAME"

    @classmethod
    def mandatory_secrets(cls):
        return [
            cls.STREAM_PROFILE_NAME,
            cls.TSDB_PROFILE_NAME,
        ]


class GetEventsFormat(MonitoringStrEnum):
    SINGLE = "single"
    SEPARATION = "separation"
    INTERSECTION = "intersection"


class FileTargetKind:
    ENDPOINTS = "endpoints"
    EVENTS = "events"
    PREDICTIONS = "predictions"
    STREAM = "stream"
    PARQUET = "parquet"
    APPS_PARQUET = "apps_parquet"
    LOG_STREAM = "log_stream"
    MONITORING_SCHEDULES = "monitoring_schedules"
    MONITORING_APPLICATION = "monitoring_application"
    ERRORS = "errors"
    STATS = "stats"
    PARQUET_STATS = "parquet_stats"
    LAST_REQUEST = "last_request"


class ModelMonitoringMode(StrEnum):
    enabled = "enabled"
    disabled = "disabled"


class ScheduleChiefFields(StrEnum):
    LAST_REQUEST = "last_request"
    LAST_ANALYZED = "last_analyzed"


class EndpointType(IntEnum):
    NODE_EP = 1  # end point that is not a child of a router
    ROUTER = 2  # endpoint that is router
    LEAF_EP = 3  # end point that is a child of a router
    BATCH_EP = 4  # endpoint that is representing an offline batch endpoint

    @classmethod
    def top_level_list(cls):
        return [cls.NODE_EP, cls.ROUTER, cls.BATCH_EP]


class EndpointMode(IntEnum):
    REAL_TIME = 0
    BATCH = 1
    BATCH_LEGACY = 2  # legacy batch mode, used for endpoints created through the batch inference job


class MonitoringFunctionNames(MonitoringStrEnum):
    STREAM = "model-monitoring-stream"
    APPLICATION_CONTROLLER = "model-monitoring-controller"
    WRITER = "model-monitoring-writer"


class V3IOTSDBTables(MonitoringStrEnum):
    APP_RESULTS = "app-results"
    METRICS = "metrics"
    EVENTS = "events"
    ERRORS = "errors"
    PREDICTIONS = "predictions"


class TDEngineSuperTables(MonitoringStrEnum):
    APP_RESULTS = "app_results"
    METRICS = "metrics"
    PREDICTIONS = "predictions"
    ERRORS = "errors"


@dataclass
class FunctionURI:
    project: str
    function: str
    tag: Optional[str] = None
    hash_key: Optional[str] = None

    @classmethod
    def from_string(cls, function_uri):
        project, uri, tag, hash_key = mlrun.common.helpers.parse_versioned_object_uri(
            function_uri
        )
        return cls(
            project=project,
            function=uri,
            tag=tag or None,
            hash_key=hash_key or None,
        )


@dataclass
class VersionedModel:
    model: str
    version: Optional[str]

    @classmethod
    def from_string(cls, model):
        try:
            model, version = model.split(":")
        except ValueError:
            model, version = model, None

        return cls(model, version)


@dataclass
class EndpointUID:
    project: str
    function: str
    function_tag: str
    function_hash_key: str
    model: str
    model_version: str
    uid: str = field(init=False)

    def __post_init__(self):
        function_ref = (
            f"{self.function}_{self.function_tag or self.function_hash_key or 'N/A'}"
        )
        versioned_model = f"{self.model}_{self.model_version or 'N/A'}"
        unique_string = f"{self.project}_{function_ref}_{versioned_model}"
        self.uid = hashlib.sha1(unique_string.encode("utf-8")).hexdigest()

    def __str__(self):
        return self.uid


class DriftStatus(Enum):
    """
    Enum for the drift status values.
    """

    NO_DRIFT = "NO_DRIFT"
    DRIFT_DETECTED = "DRIFT_DETECTED"
    POSSIBLE_DRIFT = "POSSIBLE_DRIFT"


class ResultKindApp(Enum):
    """
    Enum for the result kind values
    """

    data_drift = 0
    concept_drift = 1
    model_performance = 2
    system_performance = 3
    mm_app_anomaly = 4


class ResultStatusApp(IntEnum):
    """
    Enum for the result status values, detected means that the app detected some problem.
    """

    irrelevant = -1
    no_detection = 0
    potential_detection = 1
    detected = 2


class ModelMonitoringLabel:
    KEY = mlrun.common.constants.MLRunInternalLabels.mlrun_type
    VAL = ""

    def __str__(self) -> str:
        return f"{self.KEY}={self.VAL}"


class ModelMonitoringAppLabel(ModelMonitoringLabel):
    VAL = "mlrun__model-monitoring-application"


class ModelMonitoringInfraLabel(ModelMonitoringLabel):
    VAL = "mlrun__model-monitoring-infra"


class HistogramDataDriftApplicationConstants:
    NAME = "histogram-data-drift"
    GENERAL_RESULT_NAME = "general_drift"


class PredictionsQueryConstants:
    DEFAULT_AGGREGATION_GRANULARITY = "10m"
    INVOCATIONS = "invocations"


class SpecialApps:
    MLRUN_INFRA = "mlrun-infra"


class ModelMonitoringLabels:
    MLRUN_MODEL_MONITORING_INFRA = "mlrun-model-monitoring-infra"


_RESERVED_FUNCTION_NAMES = MonitoringFunctionNames.list() + [SpecialApps.MLRUN_INFRA]


class ModelEndpointMonitoringMetricType(StrEnum):
    RESULT = "result"
    METRIC = "metric"


# refer to `mlrun.utils.regex.project_name`
_INNER_PROJECT_PATTERN = r"[a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?"
PROJECT_PATTERN = rf"^{_INNER_PROJECT_PATTERN}$"

MODEL_ENDPOINT_ID_PATTERN = r"^[a-zA-Z0-9_-]+$"

_FQN_PART_PATTERN = r"[a-zA-Z0-9_-]+"
_RESULT_NAME_PATTERN = r"[a-zA-Z_][a-zA-Z0-9_]*"

FQN_PATTERN = (
    rf"^(?P<project>{_INNER_PROJECT_PATTERN})\."
    rf"(?P<app>{_FQN_PART_PATTERN})\."
    rf"(?P<type>{ModelEndpointMonitoringMetricType.RESULT}|{ModelEndpointMonitoringMetricType.METRIC})\."
    rf"(?P<name>{_RESULT_NAME_PATTERN})$"
)
FQN_REGEX = re.compile(FQN_PATTERN)
APP_NAME_REGEX = re.compile(_FQN_PART_PATTERN)
RESULT_NAME_REGEX = re.compile(_RESULT_NAME_PATTERN)


INTERSECT_DICT_KEYS = {
    ModelEndpointMonitoringMetricType.METRIC: "intersect_metrics",
    ModelEndpointMonitoringMetricType.RESULT: "intersect_results",
}

CRON_TRIGGER_KINDS = ("http", "cron")
STREAM_TRIGGER_KINDS = ("v3io-stream", "kafka-cluster")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import importlib.metadata as importlib_metadata
import os
import sys
import tempfile
import warnings
from types import ModuleType
from typing import Any, Optional, Union

from mlrun.errors import MLRunInvalidArgumentError
from mlrun.utils import logger


class Pickler:
    """
    A static class to pickle objects with multiple modules while capturing the environment of the pickled object. The
    pickler will raise warnings in case the object is un-pickled in a mismatching environment (different modules
    and / or python versions)
    """

    @staticmethod
    def pickle(
        obj: Any, pickle_module_name: str, output_path: Optional[str] = None
    ) -> tuple[str, dict[str, Union[str, None]]]:
        """
        Pickle an object using the given module. The pickled object will be saved to file to the given output path.

        :param obj:                The object to pickle.
        :param pickle_module_name: The pickle module to use. For example: "pickle", "joblib", "cloudpickle".
        :param output_path:        The output path to save the 'pkl' file to. If not provided, the pickle will be saved
                                   to a temporary directory. The user is responsible to clean the temporary directory.

        :return: A tuple of the path of the 'pkl' file and the instructions the pickler noted.
        """
        # Get the pickle module:
        pickle_module = importlib.import_module(pickle_module_name)
        Pickler._validate_pickle_module(pickle_module=pickle_module)
        pickle_module_version = Pickler._get_module_version(
            module_name=pickle_module_name
        )

        # Get the object's module (module name can be extracted usually from the object's class):
        object_module_name = (
            obj.__module__.split(".")[0]
            if hasattr(obj, "__module__")
            else type(obj).__module__.split(".")[0]
        )
        object_module_version = Pickler._get_module_version(
            module_name=object_module_name
        )

        # Get the python version:
        python_version = Pickler._get_python_version()

        # Construct the pickler labels dictionary (versions may not be available):
        instructions = {
            "object_module_name": object_module_name,
            "pickle_module_name": pickle_module_name,
            "python_version": python_version,
        }
        if object_module_version is not None:
            instructions["object_module_version"] = object_module_version
        if pickle_module_version is not None:
            instructions["pickle_module_version"] = pickle_module_version

        # Generate a temporary output path if not provided:
        if output_path is None:
            output_path = os.path.join(tempfile.mkdtemp(), "obj.pkl")

        # Pickle the object to file:
        with open(output_path, "wb") as pkl_file:
            pickle_module.dump(obj, pkl_file)

        return output_path, instructions

    @staticmethod
    def unpickle(
        pickle_path: str,
        pickle_module_name: str,
        object_module_name: Optional[str] = None,
        python_version: Optional[str] = None,
        pickle_module_version: Optional[str] = None,
        object_module_version: Optional[str] = None,
    ) -> Any:
        """
        Unpickle an object using the given instructions. Warnings may be raised in case any of the versions are
        mismatching (only if provided - not None).

        :param pickle_path:           Path to the 'pkl' file to un-pickle.
        :param pickle_module_name:    Module to use for unpickling the object.
        :param object_module_name:    The original object's module. Used to verify the current interpreter object module
                                      version match the pickled object version before unpickling the object.
        :param python_version:        The python version in which the original object was pickled. Used to verify the
                                      current interpreter python version match the pickled object version before
                                      unpickling the object.
        :param pickle_module_version: The pickle module version. Used to verify the current interpreter module version
                                      match the one who pickled the object before unpickling it.
        :param object_module_version: The original object's module version to match to the interpreter's module version.

        :return: The un-pickled object.
        """
        # Check the python version against the pickled object:
        if python_version is not None:
            current_python_version = Pickler._get_python_version()
            if python_version != current_python_version:
                logger.warn(
                    f"MLRun is trying to load an object that was pickled on python version "
                    f"'{python_version}' but the current python version is '{current_python_version}'. "
                    f"When using pickle, it is recommended to save and load an object on the same python version to "
                    f"reduce unexpected errors."
                )

        # Get the pickle module:
        pickle_module = importlib.import_module(pickle_module_name)
        Pickler._validate_pickle_module(pickle_module=pickle_module)

        # Check the pickle module against the pickled object (only if the version is given):
        if pickle_module_version is not None:
            current_pickle_module_version = Pickler._get_module_version(
                module_name=pickle_module_name
            )
            if pickle_module_version != current_pickle_module_version:
                logger.warn(
                    f"MLRun is trying to load an object that was pickled using "
                    f"{pickle_module_name} version {pickle_module_version} but the current module version is "
                    f"'{current_pickle_module_version}'. "
                    f"When using pickle, it is recommended to save and load an "
                    f"object using the same pickling module version to reduce unexpected errors."
                )

        # Check the object module against the pickled object (only if the version is given):
        if object_module_version is not None and object_module_name is not None:
            current_object_module_version = Pickler._get_module_version(
                module_name=object_module_name
            )
            if object_module_version != current_object_module_version:
                logger.warn(
                    f"MLRun is trying to load an object from module {object_module_name} version "
                    f"{object_module_version} but the current module version is '{current_object_module_version}'. "
                    f"When using pickle, it is recommended to save and load an object using "
                    f"the same exact module version to reduce unexpected errors."
                )

        # Load the object from the pickle file:
        with open(pickle_path, "rb") as pickle_file:
            obj = pickle_module.load(pickle_file)

        return obj

    @staticmethod
    def _validate_pickle_module(pickle_module: ModuleType):
        """
        Validate the pickle module to use have a `dump` and `load` functions so the Pickler can use it.

        :param pickle_module: The pickle module tot validate.

        :raise MLRunInvalidArgumentError: If the pickle module is not valid.
        """
        for function_name in ["dump", "load"]:
            if not hasattr(pickle_module, function_name):
                raise MLRunInvalidArgumentError(
                    f"A pickle module is expected to have a `{function_name}` function but the provided module "
                    f"{pickle_module.__name__} does not have it."
                )

    @staticmethod
    def _get_module_version(module_name: str) -> Union[str, None]:
        """
        Get a module's version. Most updated modules have versions but some don't. In case the version could not be
        read, None is returned.

        :param module_name: The module's name to get its version.

        :return: The module's version if found and None otherwise.
        """
        # First we'll try to get the module version from `importlib`:
        try:
            return importlib_metadata.version(module_name)
        except importlib.metadata.PackageNotFoundError:
            # `PackageNotFoundError` is ignored this is raised when `version` could not find the package related to the
            # module.
            pass

        # Secondly, if importlib could not get the version, we'll try to use `pkg_resources` to get the version (the
        # version will be found only if the package name is equal to the module name. For example, if the module name is
        # 'x' then the way we installed the package must be 'pip install x'):
        import pkg_resources

        with warnings.catch_warnings():
            # If a module's package is not found, a `PkgResourcesDeprecationWarning` warning will be raised and then
            # `DistributionNotFound` exception will be raised, so we ignore them both:
            warnings.filterwarnings(
                "ignore", category=pkg_resources.PkgResourcesDeprecationWarning
            )
            try:
                return pkg_resources.get_distribution(module_name).version
            except pkg_resources.DistributionNotFound:
                pass

        # The version could not be found.
        return None

    @staticmethod
    def _get_python_version() -> str:
        """
        Get the current running python's version.

        :return: The python version string.
        """
        return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
